"""Classes for tiling maps.

`weavingspace.tile_map.Tiling` and
`weavingspace.tile_map.TiledMap` are exposed in the  public API and
respectively enable creation of a tiling and plotting of the tiling as a
multivariate map.
"""

from __future__ import annotations

import copy
import itertools
import typing
from collections.abc import Iterable
from dataclasses import dataclass
from time import perf_counter

import geopandas as gpd
import matplotlib.colors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import shapely.affinity as affine
import shapely.geometry as geom
import shapely.ops

from weavingspace import Tileable, TileUnit, tiling_utils

if typing.TYPE_CHECKING:
  from matplotlib.figure import Figure

CMAPS_SEQUENTIAL = list(itertools.chain(*[[x, x + "_r"] for x in
  ["Greys", "Purples", "Blues", "Greens", "Oranges", "Reds",
   "viridis", "plasma", "inferno", "magma", "cividis",
   "YlOrBr", "YlOrRd", "OrRd", "PuRd", "RdPu", "BuPu",
   "GnBu", "PuBu", "YlGnBu", "PuBuGn", "BuGn", "YlGn",
   "binary", "gist_yarg", "gist_gray", "gray", "bone",
   "pink", "spring", "summer", "autumn", "winter", "cool",
   "Wistia", "hot", "afmhot", "gist_heat", "copper"]]))

CMAPS_DIVERGING = list(itertools.chain(*[[x, x + "_r"] for x in
  ["PiYG", "PRGn", "BrBG", "PuOr", "RdGy", "RdBu", "RdYlBu",
   "RdYlGn", "Spectral", "coolwarm", "bwr", "seismic"]]))
# these ones not yet in MPL 3.8.4: "berlin", "managua", "vanimo"

CMAPS_CATEGORICAL = list(itertools.chain(*[[x, x + "_r"] for x in
  ["Pastel1", "Pastel2", "Paired", "Accent", "Dark2",
   "Set1", "Set2", "Set3", "tab10", "tab20", "tab20b",
   "tab20c"]]))

@dataclass(slots=True, init=False)
class _TileGrid:
  """A class to represent the translation centres of a tiling.

  We store the grid as a GeoSeries of Point objects to make it simple to plot
  in map views if required.

  Implementation relies on transforming the translation vectors into a square
  space where the tile spacing is unit squares, then transforming this back
  into the original map space. Some member variables of the class are in
  the transformed grid generation space, some in the map space. We store the
  translation vectors as shapely Points, and the extent of the grid in grid
  space as a shapely Polygon so we can more easily visualise if needed.
  """

  tile_unit:Tileable
  """the base tile in map space."""
  oriented_rect_to_tile:geom.Polygon
  """oriented rectangular region in map space that is to be tiled."""
  to_grid_space:tuple[float,...]
  """the forward transformation from map space to tiling grid generation space,
  stored as a shapely.affinity transform tuple of 6 floats."""
  to_map_space:tuple[float,...]
  """the inverse transform from tiling grid space to map space."""
  extent_in_grid_space:geom.Polygon
  """geometry of the circular extent of the tiling transformed into grid
  generation space."""
  centre:geom.Point
  """centre point of the grid in map space - this is required for later
  rotations of a generated tiling in map space"""
  points:gpd.GeoSeries
  """geom.Points recording translation vectors of the tiling in map space."""

  def __init__(
      self,
      tile_unit:Tileable,
      to_tile:gpd.GeoSeries,
      at_centroids:bool = False) -> None:
    self.tile_unit = tile_unit
    self.oriented_rect_to_tile = self._get_rect_to_tile(to_tile)
    self.to_map_space, self.to_grid_space = self._get_transforms()
    self._set_centre_in_map_space()
    self._set_extent_in_grid_space()
    if at_centroids:
      self.points = to_tile.representative_point()
    else:
      self.points = self._get_grid()
    self.points.crs = self.tile_unit.crs


  def _get_rect_to_tile(
      self,
      region_to_tile:gpd.GeoSeries) -> geom.Polygon:
    """Generate an oriented rectangle that encompasses the region to be tiled.

    This will be rotated to generate a circular are that will actually be tiled
    so that if new orientations of the tiled pattern are requested they can be
    generated quickly without much recalculation.

    Args:
      region_to_tile (gpd.GeoSeries): the region to be tiled.

    Returns:
      geom.Polygon: an oriented rectangle encompassing the region to be tiled
        with some buffering to avoid cutting off tiling elements.

    """
    # buffer the region by an amount dictated by the size of the tile unit
    bb = self.tile_unit.tiles.total_bounds
    diagonal = np.hypot(bb[2] - bb[0], bb[3] - bb[1])
    return (region_to_tile.union_all().buffer(diagonal)
            .minimum_rotated_rectangle)


  def _get_transforms(self) -> tuple[tuple[float,...],tuple[float,...]]:
    """Return forward and inverse transforms from map space to grid space.

    In grid generation space the translation vectors are (1, 0) and (0, 1)
    so we simply form a matrix from two translation vectors to get the
    transform from grid space to map space, and invert it to get transform from
    map space to grid space.

    See pages 18-22 in Kaplan CS, 2009. Introductory Tiling Theory for Computer
    Graphics (Morgan & Claypool) for the logic of this approach.

    The results are returned as shapely affine transform tuples.

    Returns:
      tuple[tuple[float,...],tuple[float,...]]: shapely affine transform tuples
        (a, b, d, e, dx, dy).

    """
    v = self.tile_unit.get_vectors()
    # unpack the first two vectors and funnel them into a 2x2 matrix
    grid_to_map = np.array([[v[0][0], v[1][0]], [v[0][1], v[1][1]]])
    map_to_grid = np.linalg.inv(grid_to_map)
    return (self._np_to_shapely_transform(grid_to_map),
            self._np_to_shapely_transform(map_to_grid))


  def _set_centre_in_map_space(self) -> None:
    """Set the tiling centre in map space."""
    self.centre = self.oriented_rect_to_tile.centroid


  def _set_extent_in_grid_space(self) -> None:
    """Set extent of the grid in grid generation space."""
    corner = geom.Point(self.oriented_rect_to_tile.exterior.coords[0])
    radius = self.centre.distance(corner)
    self.extent_in_grid_space = \
      affine.affine_transform(self.centre.buffer(radius), self.to_grid_space)


  def _get_grid(self) -> gpd.GeoSeries:
    """Generate the grid transformed into map space.

    Obtain dimensions of the transformed region, then set down a uniform grid.
    Grid generation is greatly accelerated by using the numpy meshgrid method
    which we can do because we're working in a grid-generation space where
    tile units have been transformed to the unit square.

    Returns:
      gpd.GeoSeries: the grid as a collection of geom.Points.

    """
    bb = self.extent_in_grid_space.bounds
    _w, _h, _l, _b = bb[2] - bb[0], bb[3] - bb[1], bb[0], bb[1]
    w = int(np.ceil(_w))
    h = int(np.ceil(_h))
    l = _l - (w - _w) / 2
    b = _b - (h - _h) / 2
    xs, ys = np.array(np.meshgrid(np.arange(w) + l,
                                  np.arange(h) + b)).reshape((2, w * h))
    pts = [geom.Point(x, y) for x, y in zip(xs, ys, strict = True)]
    return (gpd.GeoSeries(
      [p for p in pts if p.within(self.extent_in_grid_space)])
        .affine_transform(self.to_map_space))


  def _np_to_shapely_transform(
      self,
      array:np.ndarray) -> tuple[float,...]:
    """Convert a numpy affine transform matrix to shapely format.

        [[a b c]
         [d e f] --> (a b d e c f)
         [g h i]]

    There is no translation in transforms in this use-case so c == f == 0

    Args:
      array (np.ndarray): numpy affine transform array to convert.

    Returns:
      tuple[float,...]: shapely affine transform tuple

    """
    return (*list(array[:2, :2].flatten()), 0, 0)


@dataclass(slots=True, init=False)
class Tiling:
  """Class that applies a `Tileable` object to a region to be mapped.

  The result of the tiling procedure is stored in the `tiles` variable and
  covers a region sufficient that the tiling can be rotated to any desired
  angle. Rotation can be requested when the render method is called.
  """

  tileable:Tileable
  """Tileable on which the tiling is based."""
  region:gpd.GeoDataFrame
  """the region to be tiled."""
  region_union:geom.Polygon
  """a single polygon of all the areas in the region to be tiled"""
  grid:_TileGrid
  """the grid which will be used to apply the tiling."""
  tiles:gpd.GeoDataFrame
  """the tiles after tiling has been carried out."""
  prototiles:gpd.GeoDataFrame
  """the prototiles after tiling has been carried out."""
  rotation:float
  """additional rotation applied to the tiling beyond any that might have
  been 'baked in' to the Tileable."""

  def __init__(
      self,
      tileable:Tileable,
      region:gpd.GeoDataFrame,
      as_icons:bool = False,
    ) -> None:
    """Construct a tiling by polygons extending beyond supplied region.

    The tiling is extended sufficiently to allow for its application at any
    rotation.

    Args:
      tileable (Tileable): the TileUnit or WeaveUnit to use.
      region (gpd.GeoDataFrame): the region to be tiled.
      as_icons (bool, optional): if True prototiles will only be placed at the
        region's zone centroids, one per zone. Defaults to False.

    """
    self.tileable = tileable
    self.rotation = 0
    self.region = region
    self.region.sindex # this probably speeds up overlay
    self.region_union = self.region.geometry.union_all()
    self.grid = _TileGrid(
      self.tileable,
      self.region.geometry if as_icons else gpd.GeoSeries([self.region_union]),
      as_icons)
    self.tiles, self.prototiles = self.make_tiling()
    self.tiles.sindex # again this probably speeds up overlay


  def get_tiled_map(
      self,
      rotation:float = 0.,
      join_on_prototiles:bool = False,
      retain_tileables:bool = False,
      prioritise_tiles:bool = True,
      ragged_edges:bool = True,
      use_centroid_lookup_approximation:bool = False,
      debug:bool = False,
    ) -> TiledMap:
    """Return a `TiledMap` filling a region at the requested rotation.

    HERE BE DRAGONS! This function took a lot of trial and error to get right,
    so modify with CAUTION!

    The `proritise_tiles = True` option means that the tiling will not break
    up the tiles in `TileUnit`s at the boundaries between areas in the mapped
    region, but will instead ensure that tiles remain complete, picking up
    their data from the region zone which they overlap the most.

    The exact order in which operations are performed affects performance. For
    example, the final clipping to self.region when ragged_edges = False is
    _much_ slower if it is carried out before the dissolving of tiles into the
    region zones. So... again... modify CAREFULLY!

    Args:
      rotation (float, optional): Optional rotation to apply. Defaults to 0.
      join_on_prototiles (bool, optional): if True data from the region dataset
        are joined to tiles based on the prototile to which they belong. If
        False the join is based on the tiles in relation to the region areas.
        For weave-based tilings False is probably to be preferred. Defaults to
        False.
      retain_tileables (bool, optional): if True complete tileable units will
        be retained. If False tile unit elements that do not overlap the map
        area will be discarded.
      prioritise_tiles (bool, optional): if True tiles will not be broken at
        boundaries in the region dataset. Defaults to True.
      ragged_edges (bool, optional): if True tiles at the edge of the region
        will not be cut by the region extent - ignored if prioritise_tiles is
        False when edges will always be clipped to the region extent. Defaults
        to True.
      use_centroid_lookup_approximation (bool, optional): if True use tile
        centroids for lookup of region data - ignored if prioritise_tiles is
        False when it is irrelevant. Defaults to False.
      debug (bool, optional): if True prints timing messages. Defaults to
        False.

    Returns:
      TiledMap: a TiledMap of the region with attributes attached to tiles.

    """
    if debug:
      t1 = perf_counter()

    id_var = self._setup_region_DZID()
    if join_on_prototiles:
      if rotation == 0:
        tiled_map, join_layer = self.tiles, self.prototiles
      else:
        tiled_map, join_layer = self.rotated(rotation)
      tiled_map["joinUID"] = self.tiles["prototile_id"]
    else:
      tiled_map = self.tiles if rotation == 0 else self.rotated(rotation)[0]
      tiled_map["joinUID"] = self.tiles["tile_id"]
      join_layer = tiled_map
    join_layer["joinUID"] = list(range(join_layer.shape[0]))

    # compile a list of the variable names we are NOT going to change
    # i.e. everything except the geometry and the id_var
    region_vars = [column for column in self.region.columns
                   if "geom" not in column and column != id_var]

    if debug:
      t2 = perf_counter()
      print(f"STEP 1: prep data (rotation if requested): {t2 - t1:.3f}")

    if prioritise_tiles:
      # maintain tile continuity across zone boundaries
      # so we have to do more work than a simple overlay
      if use_centroid_lookup_approximation:
        t5 = perf_counter()
        tile_pts = copy.deepcopy(join_layer)
        tile_pts.geometry = tile_pts.centroid
        lookup = tile_pts.sjoin(
          self.region, how = "inner")[["joinUID", id_var]]
      else:
        # determine areas of overlapping tiles and drop the data we join the
        # data back later, so dropping makes that easier overlaying in region.
        # overlay(tiles) seems to be faster??
        # TODO: also... this part is performance-critical, think about fixes --
        # possibly including the above centroid-based approximation
        overlaps = self.region.overlay(join_layer, make_valid = False)
        if debug:
          t3 = perf_counter()
          print(f"STEP A2: overlay zones with tiling: {t3 - t2:.3f}")
        overlaps["area"] = overlaps.geometry.area
        if debug:
          t4 = perf_counter()
          print(f"STEP A3: calculate areas: {t4 - t3:.3f}")
        overlaps.drop(columns = region_vars, inplace = True)
        if debug:
          t5 = perf_counter()
          print(f"STEP A4: drop columns prior to join: {t5 - t4:.3f}")
        # make a lookup by largest area tile to region id
        lookup = overlaps \
          .iloc[overlaps.groupby("joinUID")["area"] \
          .agg(pd.Series.idxmax)][["joinUID", id_var]]
      # now join the lookup and from there the region data
      if debug:
        t6 = perf_counter()
        print(f"STEP A5: build lookup for join: {t6 - t5:.3f}")
      tiled_map = tiled_map \
        .merge(lookup, on = "joinUID") \
        .merge(self.region.drop(columns = ["geometry"]), on = id_var)
      if debug:
        t7 = perf_counter()
        print(f"STEP A6: perform lookup join: {t7 - t6:.3f}")
      tiled_map.drop(columns = ["joinUID"], inplace = True)

    else:  
      # here it's a simple overlay
      tiled_map = self.region.overlay(tiled_map)
      t7 = perf_counter()
      if debug:
        print(f"STEP B2: overlay tiling with zones: {t7 - t2:.3f}")

    if not retain_tileables:
      tiled_map = tiled_map.loc[
        shapely.intersects(self.region_union, np.array(tiled_map.geometry)), :]

    tiled_map.drop(columns = [id_var], inplace = True)
    self.region.drop(columns = [id_var], inplace = True)

    # if we've retained tiles and want 'clean' edges, then clip
    # note that this step is slow: geopandas unary_unions the clip layer
    if prioritise_tiles and not ragged_edges:
      tiled_map.sindex
      tiled_map = tiled_map.clip(self.region_union)
      if debug:
        print(f"""STEP A7/B3: clip map to region: {perf_counter() - t7:.3f}""")

    tm = TiledMap()
    tm.tiling = self
    tm.map = tiled_map
    return tm


  def _setup_region_DZID(self) -> str:
    """Return guaranteed-unique new attribute name for self.region dataframe.

    Avoids a name clash with any existing attribute in the dataframe.

    Returns:
      str: name of the added attribute.

    """
    dzid = "DZID"
    i = 0
    while dzid in self.region.columns:
      dzid = "DZID" + str(i)
      i = i + 1
    self.region[dzid] = list(range(self.region.shape[0]))
    return dzid


  def make_tiling(self) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
    """Tile the region with a tile unit, returning a GeoDataFrame.

    Returns:
      geopandas.GeoDataFrame: a GeoDataFrame of the region tiled with the
        tile unit.

    """
    # we assume the geometry column is called geometry so make it so...
    if self.region.geometry.name != "geometry":
      self.region.rename_geometry("geometry", inplace = True)

    # chain list of lists of GeoSeries geometries to list of geometries
    tiles = itertools.chain(*[
      self.tileable.tiles.geometry.translate(p.x, p.y)
      for p in self.grid.points])
    prototiles = itertools.chain(*[
      self.tileable.prototile.geometry.translate(p.x, p.y)
      for p in self.grid.points])
    # replicate the tile ids
    tile_ids = list(self.tileable.tiles.tile_id) * len(self.grid.points)
    prototile_ids = list(range(len(self.grid.points)))
    tile_prototile_ids = sorted(prototile_ids * self.tileable.tiles.shape[0])
    tiles_gs = gpd.GeoSeries(list(tiles))
    prototiles_gs = gpd.GeoSeries(list(prototiles))
    # assemble and return as GeoDataFrames
    tiles_gdf = gpd.GeoDataFrame(
      data = {"tile_id": tile_ids, "prototile_id": tile_prototile_ids},
      geometry = tiles_gs, crs = self.tileable.crs)
    prototiles_gdf = gpd.GeoDataFrame(
      data = {"prototile_id": prototile_ids},
      geometry = prototiles_gs, crs = self.tileable.crs)
    return tiles_gdf, prototiles_gdf


  def rotated(self,
              rotation:float = 0.0,
              ) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
    """Return the stored tiling rotated.

    The stored tiling never changes and
    if it was originally made with a Tileable that was rotated it will retain
    that rotation. The requested rotation is _additional_ to that baseline
    rotation.

    Args:
      rotation (float, optional): Rotation angle in degrees.
        Defaults to None.

    Returns:
      gpd.GeoDataFrame: Rotated tiling.

    """
    if self.tiles is None:
      self.tiles = self.make_tiling()[0]
    if rotation == 0:
      return self.tiles, self.prototiles
    tiles = gpd.GeoDataFrame(
      data = {"tile_id": self.tiles.tile_id,
              "prototile_id": self.tiles.tile_id},
      crs = self.tiles.crs,
      geometry = self.tiles.geometry.rotate(
        rotation, origin = self.grid.centre))
    prototiles = gpd.GeoDataFrame(
      data = {"prototile_id": self.prototiles.prototile_id},
      crs = self.prototiles.crs,
      geometry = self.prototiles.geometry.rotate(
        rotation, origin = self.grid.centre))
    self.rotation = rotation
    return tiles, prototiles


@dataclass(slots=True)
class TiledMap:
  """Class representing a tiled map.

  Should not be accessed directly, but will
  be created by calling `Tiling.get_tiled_map()`. After creation the variables
  and colourmaps attributes can be set, and then `TiledMap.render()` called to
  make a map. Settable attributes are explained in documentation of the
  `TiledMap.render()` method.

  Examples:
    Recommended usage is as follows. First, make a `TiledMap` from a `Tiling`
    object:

      `tm = tiling.get_tiled_map(...)`

    Some options in the `Tiling` constructor affect the map appearance. See
    `Tiling` for details.

    Once a `TiledMap` object exists, set options on it, either when calling
    `TiledMap.render()` or explicitly, i.e.

      tm.render(opt1 = val1, opt2 = val2, ...)

    or

      tm.opt1 = val1
      tm.opt2 = val2
      tm.render()

    Option settings are persistent, i.e. unless a new `TiledMap` object is
    created the option settings have to be explicitly reset to new values on
    subsequent calls to `TiledMap.render()`.

    The most important options are the `vars_map` and `colors_to_use` settings.

    `vars_to_map` is a lost of the dataset variable names to match with
    `weavingspace.tileable.Tileable` elements with corresponding (ordered)
    tile_ids (usually "a", "b", etc.). If you need to control the match, then
    also supply `ids_to_map` in matching order. E.g.

      tm.ids_to_map = ['d', 'c', 'b', 'a']
      tm.vars_to_map = ['x1', 'x2', 'x3', 'x4']

    Note that this means that if you really want more than one element in the
    tiling to represent the same variable more than once, you can do that.

    `colors_to_use` is a parallel list of named matplotlib colormaps,

      tm.colors_to_use = ["Reds", "Blues", "Greys", "Purples"]

    Similarly, you can specify the classification `schemes_to_use` (such as
    'quantiles') and the number of classes `n_classes` in each.

    If data are categorical, this is flagged in the `categoricals` list of
    booleans, in which case an appropriate colour map should be used. There is
    currently no provision for control of which colour in a categorical
    colour map is applied to which variable level.

    TODO: better control of categorical mapping schemes.

  """

  # these will be set at instantion by Tiling.get_tiled_map()
  tiling:Tiling = None
  """the Tiling with the required tiles"""
  map:gpd.GeoDataFrame = None
  """the GeoDataFrame on which this map is based"""
  ids_to_map:list[str] = None
  """tile_ids that are to be used to represent data"""
  vars_to_map:list[str] = None
  """dataset variables that are to be symbolised"""
  colors_to_use:list[str|list[str]] = None
  """list of matplotlib colormap names."""
  categoricals:list[bool] = None
  """list specifying if each variable is -- or is to be treated as --
  categorical"""
  schemes_to_use:list[str|None] = None
  """mapclassify schemes to use for each variable."""
  n_classes:list[int|None] = None
  """number of classes to apply; if set to 0 will be unclassed."""
  _colourspecs:dict[str,dict] = None
  """dictionary of dictionaries keyed by the items in `ids_to_use` with each
  dictionary forming additional kwargs to be supplied to geopandas.plot()."""

  # the below parameters can be set either before calling self.render() or
  # passed in as parameters to self.render(). These are solely
  # `TiledMap.render()` options not geopandas plot options.
  legend:bool = True
  """whether or not to show a legend"""
  legend_zoom:float = 1.0
  """<1 zooms out from legend to show more context"""
  legend_dx:float = 0.
  """x shift of legend relative to the map"""
  legend_dy:float = 0.
  """y shift of legend relative to the map"""
  use_ellipse:bool = False
  """if True clips legend with an ellipse"""
  ellipse_magnification:float = 1.0
  """magnification to apply to clip ellipse"""
  radial_key:bool = False
  """if True use radial key even for ordinal/ratio data (normally these will be
  shown by concentric tile geometries)"""
  draft_mode:bool = False
  """if True plot the map coloured by tile_id"""

  # the parameters below are geopandas.plot options which we intercept to
  # ensure they are applied appropriately when we plot a GDF
  figsize:tuple[float,float] = (20, 15)
  """maptlotlib figsize"""
  dpi:float = 72
  """dpi for bitmap formats"""

  def render(
      self,
      **kwargs,
    ) -> Figure:
    """Render the current state to a map.

    Note that TiledMap objects will usually be created by calling
    `Tiling.get_tiled_map()`.

    Args:
      ids_to_map (list[str]): tile_ids to be used in the map. Defaults to None.
      vars_to_map (list[str]): dataset columns to be mapped. Defaults to None.
      colors_to_use (list[str]): list of matplotlib colormaps to be used.
        Defaults to None.
      categoricals (list[bool]): list of flags indicating if associated variable
        should be treated as categorical. Defaults to None.
      schemes_to_use (list[str]): list of strings indicating the mapclassify
        scheme to use e.g. 'equalinterval' or 'quantiles'. Defaults to None.
      n_classes (list[int]): list of ints indicating number of classes to use in
        map classification. Defaults to None.
      legend (bool): If True a legend will be drawn. Defaults to True.
      legend_zoom (float): Zoom factor to apply to the legend. Values <1
        will show more of the tile context. Defaults to 1.0.
      legend_dx (float): x shift to apply to the legend position in plot area
        relative units, i.e. 1.0 is full width of plot. Defaults to 0.0.
      legend_dy (float): x and y shift to apply to the legend position in plot
        area relative units, i.e. 1.0 is full height of plot. Defaults to 0.0.
      use_ellipse (bool): If True applies an elliptical clip to the legend.
        Defaults to False.
      ellipse_magnification (float): Magnification to apply to ellipse clipped
        legend. Defaults to 1.0.
      radial_key (bool): If True legend key for TileUnit maps will be based on
        radially dissecting the tiles, i.e. pie slices. Defaults to False.
      draft_mode (bool): If True a map of the tiled map coloured by tile_ids
        (and with no legend) is returned. Defaults to False.
      figsize (tuple[float,floar]): plot dimensions passed to geopandas.
        plot. Defaults to (20, 15).
      dpi (float): passed to pyplot.plot. Defaults to 72.
      **kwargs: other settings to pass to pyplot/geopandas.plot.

    Returns:
      matplotlib.figure.Figure: figure on which map is plotted.

    """
    plt.rcParams["pdf.fonttype"] = 42
    plt.rcParams["pdf.use14corefonts"] = True
    matplotlib.rcParams["pdf.fonttype"] = 42

    to_remove = set()  # keep track of kwargs we use to setup TiledMap
    # kwargs with no corresponding class attribute will be discarded
    # because we are using slots, we have to use setattr() here
    for k, v in kwargs.items():
      if k in self.__slots__:
        setattr(self, k, v)
        to_remove.add(k)
    # remove any them so we don't pass them on to pyplot and get errors
    for k in to_remove:
      del kwargs[k]

    if self.draft_mode:
      fig = plt.figure(figsize = self.figsize)
      ax = fig.add_subplot(111)
      self.map.plot(ax = ax, column = "tile_id", cmap = "tab20", **kwargs)
      ax.set_axis_off()
      return fig

    if self.legend:
      # this sizing stuff is rough and ready for now, possibly forever...
      reg_w, reg_h, *_ = \
        tiling_utils.get_width_height_left_bottom(self.map.geometry)
      tile_w, tile_h, *_ = \
        tiling_utils.get_width_height_left_bottom(
          self.tiling.tileable._get_legend_tiles().rotate(
            self.tiling.rotation, origin = (0, 0)))
      sf_w, sf_h = reg_w / tile_w / 3, reg_h / tile_h / 3
      gskw = {"height_ratios": [sf_h * tile_h, reg_h - sf_h * tile_h],
              "width_ratios":  [reg_w, sf_w * tile_w]}

      fig, axes = plt.subplot_mosaic([["map", "legend"], ["map", "."]],
                                        gridspec_kw = gskw,
                                        figsize = self.figsize,
                                        layout = "constrained", **kwargs)
    else:
      fig, axes = plt.subplots(1, 1, figsize = self.figsize,
                                  layout = "constrained", **kwargs)

    self._plot_map(axes, **kwargs)
    return fig


  def _plot_map(
      self,
      ax:plt.Axes,
      **kwargs,
    ) -> None:
    """Plot map to the supplied axes.

    Args:
      axes (plt.Axes): axes on which maps will be drawn.
      kwargs (dict): additional parameters to be passed to plot.

    """
    self._set_colourspecs()
    bb = self.map.geometry.total_bounds
    if self.legend:
      if (self.legend_dx != 0 or self.legend_dx != 0):
        box = ax["legend"].get_position()
        box.x0 += self.legend_dx
        box.x1 += self.legend_dx
        box.y0 += self.legend_dy
        box.y1 += self.legend_dy
        ax["legend"].set_position(box)
      ax["map"].set_axis_off()
      ax["map"].set_xlim(bb[0], bb[2])
      ax["map"].set_ylim(bb[1], bb[3])
      self._plot_subsetted_gdf(ax["map"], self.map, **kwargs)
      self.plot_legend(ax = ax["legend"], **kwargs)
    else:
      ax.set_axis_off()
      ax.set_xlim(bb[0], bb[2])
      ax.set_ylim(bb[1], bb[3])
      self._plot_subsetted_gdf(ax, self.map, **kwargs)


  def _plot_subsetted_gdf(
      self,
      ax:plt.Axes,
      gdf:gpd.GeoDataFrame,
      grouping_var:str = "tile_id",
      **kwargs,
    ) -> None:
    """Repeatedly plot a gpd.GeoDataFrame based on a subsetting attribute.

    NOTE: used to plot both the main map _and_ the legend, which is why
    a separate GeoDataframe is supplied and we don't just use self.map.

    Args:
      ax (pyplot.Axes): axes to plot to.
      gdf (gpd.GeoDataFrame): the GeoDataFrame to plot.

    Raises:
      Exception: if self.colourmaps cannot be parsed exception is raised.

    """
    groups = gdf.groupby(grouping_var)
    for ID, cspec in self._colourspecs.items():
      subset = groups.get_group(ID)
      n_values = len(subset[cspec["column"]].unique())
      if not cspec["categorical"] and n_values == 1:
        print(f"""
              Only one level in variable {cspec['column']}, replacing requested
              colour map with single colour fill.
              """)
        cspec["color"] = matplotlib.colormaps.get(cspec["cmap"])(0.5)
        del cspec["column"]
        del cspec["cmap"]
        del cspec["scheme"]
      elif not cspec["categorical"] and n_values < cspec["k"]:
        cspec["k"] = n_values
      elif cspec["categorical"]:
        del cspec["scheme"]
      subset.plot(ax = ax, **cspec, **kwargs)


  def to_file(self, fname:str) -> None:
    """Output the tiled map to a layered GPKG file.

    Currently delegates to `weavingspace.tiling_utils.write_map_to_layers()`.

    Args:
      fname (str): Filename to write. Defaults to None.

    """
    tiling_utils.write_map_to_layers(self.map, fname)


  def plot_legend(self, ax, **kwargs) -> None:
    """Plot a legend for this tiled map.

    Args:
      ax (pyplot.Axes, optional): axes to draw legend. Defaults to None.

    """
    ax.set_axis_off()
    legend_tiles = self.tiling.tileable._get_legend_tiles()
    # this is a bit hacky, but we will apply the rotation to text
    # annotation so for TileUnits which don't need it, reverse that now
    if isinstance(self.tiling.tileable, TileUnit):
      # note that this confuses type hinting because pandas silently assigns 
      # a scalar value to a Series
      legend_tiles.rotation = -self.tiling.rotation

    legend_key = self._get_legend_key_gdf(legend_tiles)
    legend_tiles.geometry = legend_tiles.geometry.rotate(
      self.tiling.rotation, origin = (0, 0))

    if self.use_ellipse:
      ellipse = tiling_utils.get_bounding_ellipse(
        legend_tiles.geometry, mag = self.ellipse_magnification)
      bb = ellipse.total_bounds
      c = ellipse.union_all().centroid
    else:
      bb = legend_tiles.geometry.total_bounds
      c = legend_tiles.geometry.union_all().centroid

    # apply legend zoom - NOTE that this must be applied even
    # if self.legend_zoom is == 1...
    ax.set_xlim(c.x + (bb[0] - c.x) / self.legend_zoom,
                c.x + (bb[2] - c.x) / self.legend_zoom)
    ax.set_ylim(c.y + (bb[1] - c.y) / self.legend_zoom,
                c.y + (bb[3] - c.y) / self.legend_zoom)

    for cs, tile, rotn in zip(self._colourspecs.values(),
                              legend_tiles.geometry,
                              legend_tiles.rotation):
      c = tile.centroid
      ax.annotate(cs["column"], xy = (c.x, c.y),
                  ha = "center", va = "center",
                  rotation_mode = "anchor",
                  # adjust rotation to favour text reading left to right
                  rotation = (rotn + self.tiling.rotation + 90) % 180 - 90,
                  bbox = {"lw": 0, "fc": "#ffffff60"})

    # now plot background; we include the central tiles, since in
    # the weave case these may not match the legend tiles
    context_tiles = self.tiling.tileable \
      .get_local_patch(r = 2, include_0 = True) \
      .geometry.rotate(self.tiling.rotation, origin = (0, 0))
    if self.use_ellipse:
      context_tiles.clip(ellipse, keep_geom_type = False).plot(
        ax = ax, fc = "#9F9F9F3F", lw = .35)
      tiling_utils.get_tiling_edges(context_tiles.geometry).clip(
        ellipse, keep_geom_type = True).plot(
          ax = ax, ec = "#5F5F5F", lw = .35)
    else:
      context_tiles.plot(ax = ax, fc = "#9F9F9F3F",
                         ec = "#5F5F5F", lw = .35)
      tiling_utils.get_tiling_edges(context_tiles.geometry).plot(
        ax = ax, ec = "#5F5F5F", lw = .35)

    # plot the legend key tiles (which include the data)
    self._plot_subsetted_gdf(ax, legend_key, **kwargs)


  def _get_legend_key_gdf(self, tiles:gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Return tiles dissected and with data assigned for use as a legend.

    'Dissection' is handled differently by `WeaveUnit` and `TileUnit`
    objects and delegated to either `WeaveUnit._get_legend_key_shapes()`
    or `TileUnit._get_legend_key_shapes()`.

    Args:
      tiles (gpd.GeoDataFrame): the legend tiles.

    Returns:
      gpd.GeoDataFrame:  with tile_id, variables and rotation
        attributes, and geometries of Tileable tiles sliced into a
        colour ramp or set of nested tiles.

    """
    key_tiles = []   # set of tiles to form a colour key (e.g. a ramp)
    ids = []         # tile_ids applied to the keys
    unique_ids = []  # list of each tile_id used in order
    vals = []        # the data assigned to the key tiles
    rots = []        # rotation of each key tile
    # subsets = self.map.groupby("tile_id")
    for (id, cspec), geom, rot in zip(self._colourspecs.items(),
                                      tiles.geometry, 
                                      tiles.rotation):
      d = list(self.map.loc[self.map.tile_id == id][cspec["column"]])
      # if the data are categorical then it's complicated...
      # if cs["categorical"]:
      #   radial = True and self.radial_key
      #   # desired order of categorical variable is the
      #   # color maps dictionary keys
      #   num_cats = len(cmap)
      #   val_order = dict(zip(cmap.keys(), range(num_cats)))
      #   # compile counts of each category
      #   freqs = [0] * num_cats
      #   for v in list(d):
      #     freqs[val_order[v]] += 1
      #   # make list of the categories containing appropriate
      #   # counts of each in the order needed using a reverse lookup
      #   data_vals = list(val_order.keys())
      #   data_vals = [data_vals[i] for i, f in enumerate(freqs) if f > 0]
      # else: # any other data is easy!
      #   data_vals = sorted(d)
      #   freqs = [1] * len(data_vals)
      data_vals = sorted(d)
      key = self.tiling.tileable._get_legend_key_shapes(          # type: ignore
        geom, [1] * len(data_vals), rot, False)
      key_tiles.extend(key)
      vals.extend(data_vals)
      n = len(data_vals)
      ids.extend([id] * n)
      unique_ids.append(id)
      rots.extend([rot] * n)
    # finally make up a data table with all the data in all the columns. This
    # allows us to reuse the tiling_utils.plot_subsetted_gdf() function. To be
    # clear: all the data from all variables are added in all columns but when
    # sent for plotting only the subset associated with each tile_id will get
    # plotted. It's wasteful of space... but note that the same is true of the
    # original data - each tile_id has data for all the variables even if it's
    # not being used to plot them: tables gonna table!
    key_data = {}
    for ID in unique_ids:
      key_data[self.vars_to_map[self.ids_to_map.index(ID)]] = vals
    key_gdf = gpd.GeoDataFrame(
      data = key_data | {"tile_id": ids, "rotation": rots},
      crs = self.map.crs,
      geometry = gpd.GeoSeries(key_tiles))
    key_gdf.geometry = key_gdf.rotate(self.tiling.rotation, origin = (0, 0))
    return key_gdf


  def explore(self) -> None:
    """TODO: add wrapper to make tiled web map via geopandas.explore.
    """
    return None


  def _set_colourspecs(self) -> None:
    """Set _colourspecs dictionary based on instance member variables.

    Set up is to the extent it is possible to follow user requested variables.
    Each requested `ids_to_map` item keys a dictionary in `_colourspecs` which
    contains the `column`, `cmap`, `scheme`, `categorical`, and `k` parameters
    to be passed on for use by the `geopandas.GeoDataFrame.plot()` calls in the
    _plot_subsetted_gdf()` method.

    This is the place to make 'smart' adjustments to how user requests for map
    styling are handled.
    """
    numeric_columns = list(self.map.select_dtypes(
      include = ("float", "int")).columns)
    # note that some numeric columns can be considered categorical
    categorical_columns = list(self.map.select_dtypes(
      include = ("category", "int")).columns)
    try:
      if isinstance(self.ids_to_map, str):
        # wrap a single string in a list - this would be an unusual request...
        if self.ids_to_map in list(self.map.tile_id):
          print("""You have only requested a single attribute to map.
                That's fine, but perhaps not what you intended?""")
          self.ids_to_map = [self.ids_to_map]
        else:
          raise KeyError(
            """You have requested a single non-existent attribute to map!""")
      elif self.ids_to_map is None or not isinstance(self.ids_to_map, Iterable):
        # default to using all of them in order
        print("""No tile ids provided: setting all of them!""")
        self.ids_to_map = sorted(list(set(self.map.tile_id)))

      if self.vars_to_map is None or not isinstance(self.vars_to_map, Iterable):
        self.vars_to_map = []
        if len(numeric_columns) == 0:
          # if there are none then we can't do it
          raise IndexError("""Attempting to set default variables, but
                          there are no numeric columns in the data!""")
        if len(numeric_columns) < len(self.ids_to_map):
          # if there are fewer available than we need then repeat some
          print("""Fewer numeric columns in the data than elements in the 
                tile unit. Reusing as many as needed to make up the numbers""")
          reps = len(self.ids_to_map) // len(numeric_columns) + 1
          self.vars_to_map = (numeric_columns * reps)[:len(self.ids_to_map)]
        elif len(numeric_columns) > len(self.ids_to_map):
          # if there are more than we need let the user know, but trim list
          print("""Note that you have supplied more variables to map than 
                there are distinct elements in the tile unit. Ignoring the
                extras.""")
          self.vars_to_map = numeric_columns[:len(self.ids_to_map)]
        else:
          self.vars_to_map = numeric_columns
      # print(f"{self.vars_to_map=}")

      if self.categoricals is None or not isinstance(self.categoricals, Iterable):
        # provide a set of defaults
        self.categoricals = [col not in numeric_columns for col in self.vars_to_map]
      # print(f"{self.categoricals=}")
      
      if isinstance(self.schemes_to_use, str):
        self.schemes_to_use = [self.schemes_to_use] * len(self.ids_to_map)
      elif self.schemes_to_use is None or not isinstance(self.schemes_to_use, Iterable):
        # provide a set of defaults
        self.schemes_to_use = [None if cat else "EqualInterval"
                              for cat in self.categoricals]
      # print(f"{self.schemes_to_use=}")
      
      if isinstance(self.colors_to_use, str):
        self.colors_to_use = [self.colors_to_use] * len(self.ids_to_map)
      elif self.colors_to_use is None or not isinstance(self.colors_to_use, Iterable):
        # provide starter defaults
        print(f"""No colour maps provided! Setting some defaults.""")
        self.colors_to_use = ["Set1" if cat else "Reds" for cat in self.categoricals]
      for i, (col, cat) in enumerate(zip(self.colors_to_use, self.categoricals)):
        if cat and col not in CMAPS_CATEGORICAL:
          self.colors_to_use[i] = CMAPS_CATEGORICAL[i]
        # we'll allow diverging schemes for now...
        elif not cat and col not in CMAPS_SEQUENTIAL and col not in CMAPS_DIVERGING:
          self.colors_to_use[i] = CMAPS_SEQUENTIAL[i]
      # print(f"{self.colors_to_use=}")

      if isinstance(self.n_classes, int):
        if self.n_classes == 0:
          self.n_classes = [255] * len(self.ids_to_map)
        else:
          self.n_classes = [self.n_classes] * len(self.ids_to_map)
      elif self.n_classes is None or not isinstance(self.n_classes, Iterable):
        # provide a set of defaults
        self.n_classes = [None if cat else 100 for cat in self.categoricals] 
      # print(f"{self.n_classes=}")

    except IndexError as e:
      e.add_note("""One or more of the supplied lists of mapping settings is 
                 an inappropriate length""")
      raise

    self._colourspecs = {
      ID: {"column": v,
           "cmap": c,
           "categorical": cat,
           "scheme": s,
           "k": k}
      for ID, v, c, cat, s, k 
      in zip(self.ids_to_map,
             self.vars_to_map,
             self.colors_to_use,
             self.categoricals,
             self.schemes_to_use,
             self.n_classes, strict = False)}



"""Functions for setting up a TileUnit with various tile geometries.

Some care is required in adding new functions that use existing ones to get the
sequence of setup operations right. Modify with care!

Many of the available tilings can be viewed at

https://dosull.github.io/mapweaver/app/

These tilings (and many many more!) are discussed in

Grunbaum B, Shephard G C, 1987 _Tilings and Patterns_ (W. H. Freeman and
Company, New York)

A more accessible introduction is

Kaplan C S, 2002 _Computer Graphics and Geometric Ornamental Design_, PhD
thesis, University of Washington, Seattle, WA,
https://cs.uwaterloo.ca/~csk/other/phd/kaplan_diss_full_print.pdf

and a more 'polished' version of that work focused on computer graphics is also
available

Kaplan C S, 2009 _Introductory Tiling Theory for Computer Graphics_ (Morgan &
Claypool)
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
  from weavingspace import TileUnit
import itertools
import math
import string

import geopandas as gpd
import numpy as np
import shapely.affinity as affine
import shapely.geometry as geom
from shapely import line_interpolate_point

from weavingspace import TileShape, tiling_utils

sqrt2 = np.sqrt(2)
sqrt3 = np.sqrt(3)

def _setup_base_tiling(unit:TileUnit) -> str|None:
  """Set up hexagonal, square, diamond, or triangular tileable unit.

  Args:
      unit (TileUnit): contains the specification requested by calling context.

  Returns:
      str|None: str if there is a problem, None otherwise.

  """
  unit.spacing = 1000 if unit.spacing is None else unit.spacing
  match unit.base_shape:
    case TileShape.RECTANGLE:
      polys = [tiling_utils.get_regular_polygon(unit.spacing, 4)]
      unit.setup_vectors((0, unit.spacing), (unit.spacing, 0))
    case TileShape.TRIANGLE:
      tri1 = tiling_utils.get_regular_polygon(unit.spacing, 3)
      tri1 = affine.translate(tri1, 0, -tri1.bounds[1])
      tri2 = affine.rotate(tri1, 180, (0, 0))
      polys = [tri1, tri2]
      unit.setup_vectors((unit.spacing / sqrt3,  unit.spacing),
                         (unit.spacing / sqrt3, -unit.spacing))
    case TileShape.DIAMOND:
      half_tile = tiling_utils.get_regular_polygon(unit.spacing, 3)
      half_tile = affine.translate(half_tile, 0, -half_tile.bounds[1])
      corners = tiling_utils.get_corners(half_tile, repeat_first = False)
      coords = [(p.x, p.y) for p in corners]
      polys = [geom.Polygon([coords[0], coords[1], coords[2],
                             (coords[1][0], -coords[1][1])])]
      unit.setup_vectors((unit.spacing / sqrt3,  unit.spacing),
                         (unit.spacing / sqrt3, -unit.spacing))
    case TileShape.HEXAGON:
      polys = [tiling_utils.get_regular_polygon(unit.spacing, 6)]
      unit.setup_vectors((0,  unit.spacing),
                         (unit.spacing * sqrt3 / 2,  unit.spacing / 2),
                         (unit.spacing * sqrt3 / 2, -unit.spacing / 2))
    case _:
      return "No proper TileShape provided to base tile unit constructor."
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("ab")[:len(polys)]},
    geometry = gpd.GeoSeries(polys),
    crs = 3857 if unit.crs is None else unit.crs)
  return None


def _setup_cairo(unit:TileUnit) -> None:
  """Set up the Cairo tiling. King of tilings. All hail the Cairo tiling.

  The pattern in this function shows how any new setup function should be
  organised:

    + Make the geometries and put them in a GeoDataFrame with the
      appropriate number of tile_id codes
    + Set the translation vectors
    + Return None (THIS IS IMPORTANT)

  This example shows this for a 'handcoded' set of geometries can be applied.

  Note that it is advisable to avoid intersection and union operations where
  possible, as it often yields floating point mismatches that can be hard to
  repair! (This tiling could be relatively conveniently generated by dissecting
  a square in 4 quarters at a 30 degree angle to the sides and then reflecting
  and rotating copies and joing them back together. But floating point issues
  make that very messy indeed. Much better to make the geometries 'pure'.)

  To design new tilings, you have to get out pencil and paper and do the
  trigonometry.

  Args:
    unit (TileUnit):  the TileUnit to setup.

  """
  # MAKE GEOMETRIES
  d = unit.spacing
  x = d/2 / (np.cos(np.pi/12) + np.cos(5*np.pi/12))
  # the following is just the geometry, it is what it is...
  # points are (more or less)
  #
  #         3
  #     2
  #            4
  #     1    0
  #
  # then rotate -15 and make 4 copies at 90 degree rotations
  p1 = geom.Polygon([(x, 0), (0, 0), (0, x),
                     (x * sqrt3/2, x + x/2),
                     (x * (1 + sqrt3) / 2, x * (3 - sqrt3) / 2)])
  p1 = affine.rotate(p1, -15, (0, 0))
  p2 = affine.rotate(p1, 90, (0, 0))
  p3 = affine.rotate(p1, 180, (0, 0))
  p4 = affine.rotate(p1, 270, (0, 0))
  # now move them so they are arranged as a hexagon centered on the tile
  p1 = affine.translate(p1, -unit.spacing / 2, 0)
  p2 = affine.translate(p2, unit.spacing / 2, 0)
  p3 = affine.translate(p3, unit.spacing / 2, 0)
  p4 = affine.translate(p4, -unit.spacing / 2, 0)
  polys = [affine.rotate(p, 45, origin = (0, 0)) for p in [p1, p2, p3, p4]]
  # MAKE THE GEODATAFRAME
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcd")}, crs = unit.crs,
    geometry = gpd.GeoSeries(polys))

  # SET THE TRANSLATION VECTORS
  unit.setup_vectors((unit.spacing / sqrt2,  unit.spacing/sqrt2),
                     (unit.spacing / sqrt2, -unit.spacing/sqrt2))

  # RETURN NONE - not strictly necessary, if all is well


def _setup_hex_slice(unit:TileUnit) -> str|None:
  """Set up tile unit with number of radial slices of a hexagon.

  Slices will start at a corner, but may be offset from there by an optional
  offset starting point.

  The supplied unit should have offset and n set.

  Places unit.n points equally spaced around the perimeter of the hexagon.
  Slices are formed by connecting some sequence of these points (and any
  intervening hexagon corners) and the hexagon centre to form a pie slice.
  unit.offset = 0 starts from corner 0. unit.offset = 1 starts half way between
  corner 0 and where corner 1 would be if offset were 0. Other values of
  unit.offset scale linearly between these.

  Args:
    unit (TileUnit):  the TileUnit to setup.

  """
  if any(k not in unit.__dict__ for k in ["n", "offset"]):
    return ("""Hex slice tiling requires both n and offset values
            to be supplied.""")
  hexagon = tiling_utils.get_regular_polygon(unit.spacing, 6)
  slices = _get_radially_sliced_polygon(hexagon, unit.n, unit.offset)
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:unit.n]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(slices))
  unit.setup_vectors((                     0,  unit.spacing    ),
                     (unit.spacing * sqrt3/2,  unit.spacing / 2),
                     (unit.spacing * sqrt3/2, -unit.spacing / 2))
  return None

def _setup_square_slice(unit:TileUnit) -> str|None:
  """Set up tile unit with number of radial slices of a square.

  The supplied unit should have offset and n set.

  Places unit.n points equally spaced around the perimeter of the square.
  Slices are formed by connecting some sequence of these points (and any
  intervening squares corners) and the square centre to form a pie slice.
  unit.offset = 0 starts from corner 0. unit.offset starts half way between
  corner 0 and where corner 1 would be if offset were 0. Other values of
  unit.offset scale linearly between these.

  Args:
    unit (TileUnit): the TileUnit to setup.

  """
  if any(k not in unit.__dict__ for k in ["n", "offset"]):
    return ("""Square slice tiling requires both n and offset values
            to be supplied.""")
  square = tiling_utils.get_regular_polygon(unit.spacing, 4)
  slices = _get_radially_sliced_polygon(square, unit.n, unit.offset)
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:unit.n]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(slices))
  unit.setup_vectors((0, unit.spacing), (unit.spacing, 0))
  return None

def _get_radially_sliced_polygon(
    shape:geom.Polygon,
    n_slices:int,
    offset:float,
  ) -> gpd.GeoSeries:
  """Return a number of 'pie slices' of the supplied polygon as a GeoSeries.

  Args:
    shape (geom.Polygon): the polygon to slice up.
    n_slices (int): the desired number of slices.
    offset (float): a linearly scaled positional offset for the starting
      point of the slices. 0 is at the first vertex. 1 is halfway between
      that and the next vertex.

  Returns:
      gpd.GeoSeries: a set of slice polygons.

  """
  n_sides = len(shape.exterior.coords) - 1
  boundary = geom.LineString(shape.exterior.coords)
  offset = (offset % 1
            if offset < 0 or offset > 1
            else offset)
  offset = (offset * n_sides / n_slices / 2
            if n_slices >= n_sides
            else offset / 2)
  # the values in along are denominated in 'side counts' i.e. they run from
  # 0 to the number of sides, with e.g., 2,5 halfway along the 3rd side
  along = [n_sides * i / n_slices + offset for i in range(n_slices)]
  along = [*along, n_sides + along[0]]
  # the cut points
  points = [line_interpolate_point(
    boundary, (a % n_sides) / n_sides, normalized = True) for a in along]
  # the corners of the original polygon
  corners = tiling_utils.get_corners(shape)
  slices = []
  # now we iterate over the list of values in along and in points in parallel
  # checking if any cut point has 'skipped' a corner when the integer part
  # of the values from the along list will be different
  for a1, a2, p1, p2 in zip(along[:-1], along[1:],
                            points[:-1], points[1:], strict = True):
    if math.floor(a1) == math.floor(a2):
      # two cut points on same polygon edge, so make the slice
      slices.append(geom.Polygon([p1, p2, (0, 0)]))
    else:
      # there is at least one corner between the cut points so get them
      skipped_corners = corners[math.ceil(a1):math.ceil(a2)]
      # but discard any that are close to the cut points...
      skipped_corners = [
        c for c in skipped_corners
        if not (np.isclose(c.distance(p1), 0, rtol = 1e-6, atol = 1e-6)
             or np.isclose(c.distance(p2), 0, rtol = 1e-6, atol = 1e-6))]
      slices.append(geom.Polygon([p1, *skipped_corners, p2, (0, 0)]))
  return gpd.GeoSeries(slices)


def _setup_crosses(unit:TileUnit) -> str|None:
  """Tilings by varying numbers of crosses.

  The supplied unit should have n set. The specific arrangements used are
  more or less arbitrary, worked out on inkscape and/or with pencil and paper.
  There is probably an algorithmic way to relate the arrangement of crosses
  and the tiling translation vectors, but this is fine for present purposes.

  Args:
    unit (TileUnit):  the TileUnit to setup.

  """
  if any(k not in unit.__dict__ for k in ["n"]):
    return ("""Crosses tiling requires n to be supplied.""")
  x = unit.spacing / np.sqrt(unit.n * 5)
  cross = geom.Polygon([
      (-x/2, -x/2), (-3*x/2,   -x/2), (-3*x/2,    x/2),
      (-x/2,  x/2), (  -x/2,  3*x/2), (   x/2,  3*x/2),
      ( x/2,  x/2), ( 3*x/2,    x/2), ( 3*x/2,   -x/2),
      ( x/2, -x/2), (   x/2, -3*x/2), (  -x/2, -3*x/2)])
  match unit.n:
    case 2:
      trans = [(-x/2, -x), (x/2, x)]
      parts = [affine.translate(cross, dx, dy) for dx, dy in trans]
      unit.setup_vectors((-3*x, -x), (-x, 3*x))
    case 3:
      trans = [(-3*x/2, -x), (-x/2, x), (3*x/2, 0)]
      parts = [affine.translate(cross, dx, dy) for dx, dy in trans]
      unit.setup_vectors((-4*x, -3*x), (-5*x, 0), (-x,  3*x))
    case 4:
      trans = [(3*x/2, x/2), (x/2, -3*x/2), (-3*x/2, -x/2), (-x/2, 3*x/2)]
      parts = [affine.translate(cross, dx, dy) for dx, dy in trans]
      unit.setup_vectors((-4*x, -3*x), (-4*x,  2*x), (0,  5*x))
    case 5:
      trans = [(0, 0), (x, 2*x), (2*x, -x), (-x, -2*x), (-2*x, x)]
      parts = [affine.translate(cross, dx, dy) for dx, dy in trans]
      unit.setup_vectors((-5*x, 0), (0,  5*x))
    case 6:
      trans = [(-2*x, -3*x/2), (-x,  x/2), (0,  5*x/2),
               ( 2*x,  3*x/2), ( x, -x/2), (0, -5*x/2)]
      parts = [affine.translate(cross, dx, dy) for dx, dy in trans]
      unit.setup_vectors((-3*x, -6*x), (-5*x, 0), (-2*x,  6*x))
    case 7:
      trans = [(0, 0), (-x, -2*x), (-2*x,  x), (-x,  3*x),
                       ( x,  2*x), ( 2*x, -x), ( x, -3*x)]
      parts = [affine.translate(cross, dx, dy) for dx, dy in trans]
      unit.setup_vectors((-4*x, -3*x), (-5*x,  5*x), (-x,  8*x))
    case _:
      return (f"""a {unit.n} crosses tiling is not implemented.
              Try a number between 2 and 7.""")
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:unit.n]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(parts))
  return None

def _setup_laves(unit:TileUnit) -> str|None:
  """Set up Laves tilings.

  See:
  https://en.wikipedia.org/wiki/List_of_Euclidean_uniform_tilings#Laves_tilings.

  These are all isohedral, but mostly not regular polygons. We prioritise them
  over the Archimedean tilings because being isohedral all tiles are the same
  size and shape which makes code to generate them a lot easier to write.
  Several are hex dissections and setup is delegated accordingly.

  Args:
    unit (TileUnit):  the TileUnit to setup.

  """
  if any(k not in unit.__dict__ for k in ["code"]):
    return ("""Laves tiling requires vertex degree code to be supplied.""")
  match unit.code:
    case "3.3.3.3.3.3":
      # this is the regular hexagons
      unit.base_shape = TileShape.HEXAGON
      return ("""Laves 6.6.6 is a basic hexagonal tiling. Perhaps you want a
              hex-colouring, hex-slice or hex-dissection tiling?""")
    case "3.3.3.3.6":
      # this one needs its own code
      _setup_laves_3_3_3_3_6(unit)
    case "3.3.3.4.4":
      # this is a boring tiling...
      _setup_archimedean_3_3_3_4_4(unit)
      unit.tiles = tiling_utils.get_dual_tile_unit(unit)
      unit.tiles.geometry = gpd.GeoSeries(
        [affine.translate(p, -unit.spacing / np.sqrt(1 + sqrt3) / 2, 0)
         for p in unit.tiles.geometry])
    case "3.3.4.3.4":
      # king of tilings!
      _setup_cairo(unit)
    case "3.4.6.4":
      # the hex 6-dissection
      unit.n = 6
      unit.offset = 1
      _setup_hex_slice(unit)
    case "3.6.3.6":
      # hex 3-dissection (also a cube weave!)
      unit.n = 3
      unit.offset = 0
      _setup_hex_slice(unit)
    case "3.12.12":
      # again this one needs its own
      _setup_laves_3_12_12(unit)
    case "4.4.4.4":
      # square grid
      unit.base_shape = TileShape.RECTANGLE
      return ("""Laves 4.4.4.4 is a basic square tiling. Perhaps you want a
              square-colouring or square-slice tiling?""")
    case "4.6.12":
      # hex 12-dissection
      unit.n = 12
      unit.offset = 0
      _setup_hex_slice(unit)
    case "4.8.8":
      # this one needs its own (a 4-dissection of the square)
      # perhaps to be added as a category later...
      unit.n = 4
      unit.offset = 0.0
      _setup_square_slice(unit)
    case "6.6.6":
      # triangles
      unit.base_shape = TileShape.TRIANGLE
      return """Note that Laves 3.3.3.3.3.3 is a basic triangle tiling."""
    case _:
      return  "[{unit.code}] is not a valid Laves code."
  return None


def _setup_laves_3_3_3_3_6(unit:TileUnit) -> None:
  """Set up Laves [3.3.3.3.6] which is like a 6 petal flower.

  Similar to the H3 7 hexagon group with the central hex removed and each
  hex 'taking' a 1/6 share of the central hex.

  Args:
    unit (TileUnit):  the TileUnit to setup.

  """
  sf = 1 / np.sqrt(7)
  tile = tiling_utils.get_regular_polygon(unit.spacing, 6)
  hexagon = affine.scale(tile, sf, sf)
  # translate it up by its own height
  hexagon = affine.translate(hexagon, 0, hexagon.bounds[3] - hexagon.bounds[1])
  hex_p = list(hexagon.exterior.coords)
  # now replace first and last points by (0, 0)
  # (note shapely doesn't require closure of the polygon)
  petal = geom.Polygon([(0, 0)] + hex_p[1:5])
  offset_a = np.degrees(np.arctan(1 / 3 / sqrt3))
  petals = [
    affine.rotate(petal, a + offset_a, origin = (0, 0))
    for a in range(30, 360, 60)]
  unit.setup_vectors((                     0,  unit.spacing    ),
                     (unit.spacing * sqrt3/2,  unit.spacing / 2),
                     (unit.spacing * sqrt3/2, -unit.spacing / 2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcdef")},
    crs = unit.crs,
    geometry = gpd.GeoSeries(petals))


def _setup_laves_3_12_12(unit:TileUnit) -> None:
  """Set up Laves [3.12.12] - like [3.6.3.6] with rhombuses sliced lengthwise.

  Args:
    unit (TileUnit):  the TileUnit to setup.

  """
  hexagon = tiling_utils.get_regular_polygon(unit.spacing, 6)
  pts = list(hexagon.exterior.coords)
  tri1 = geom.Polygon([pts[0], pts[2], [0, 0]])
  tri2 = geom.Polygon([pts[0], pts[1], pts[2]])
  tris1 = [affine.rotate(tri1, a, origin = (0, 0))
       for a in range(0, 360, 120)]
  tris2 = [affine.rotate(tri2, a, origin = (0, 0))
       for a in range(0, 360, 120)]
  # reorder so the 'inner' and 'outer' triangles are labelled alternately
  tris = itertools.chain(*zip(tris1, tris2, strict = True))
  unit.setup_vectors((                     0,  unit.spacing    ),
                     (unit.spacing * sqrt3/2,  unit.spacing / 2),
                     (unit.spacing * sqrt3/2, -unit.spacing / 2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcdef")}, crs = unit.crs,
    geometry = gpd.GeoSeries(tris))


def _setup_archimedean(unit:TileUnit) -> str|None:
  r"""Set up Archimedean 'regular' tilings.

  See:
  https://en.wikipedia.org/wiki/Euclidean_tilings_by_convex_regular_polygons

  Many of these could be constructed as duals of the Laves tilings. However,
  inaccuracies in the `tiling_utils.get_dual_tiling()` function yield tiles
  that don't sustain reliable Topology detection... ¯\_(ツ)_/¯

  Args:
    unit (TileUnit):  the TileUnit to setup.

  """
  if any(k not in unit.__dict__ for k in ["code"]):
    return "Archimedean tiling requires vertex degree code to be supplied."
  match unit.code:
    case "3.3.3.3.3.3":
      unit.base_shape = TileShape.TRIANGLE
      return "Note that Archimedean 3.3.3.3.3.3 is a basic triangle tiling?"
    case "3.3.3.3.6":
      _setup_archimedean_3_3_3_3_6(unit)
    case "3.3.3.4.4":
      _setup_archimedean_3_3_3_4_4(unit)
    case "3.3.4.3.4":
      _setup_archimedean_3_3_4_3_4(unit)
    case "3.4.6.4":
      _setup_archimedean_3_4_6_4(unit)
    case "3.6.3.6":
      _setup_archimedean_3_6_3_6(unit)
    case "3.12.12":
      _setup_archimedean_3_12_12(unit)
    case "4.4.4.4":
      unit.base_shape = TileShape.RECTANGLE
      return ("""Archimedean 4.4.4.4 is a basic square tiling. Perhaps you
              want a square-colouring or square-slice tiling?""")
    case "4.6.12":
      _setup_archimedean_4_6_12(unit)
    case "4.8.8":
      _setup_archimedean_4_8_8(unit)
    case "6.6.6":
      unit.base_shape = TileShape.HEXAGON
      return ("""Archimedean 6.6.6 is a basic hexagonal tiling. Perhaps you
              want a hex-colouring, hex-slice or hex-dissection tiling?""")
    case _:
      unit.base_shape = TileShape.RECTANGLE
      return  f"({unit.code}) is not a valid Archimedean code."
  return None


# Note that where possible in the functions that follow the following short
# variable names are used:
#  t for height of triangles
#  s for height of squares
#  h for height of hexagons
#  o for height of octagons
#  d for height of dodecagons
# rescaling so that the resulting tile unit meets the unit.spacing request is
# in most cases exact, but where the tile unit departs from a plain square or
# hexagon it's approximate.
def _setup_archimedean_3_3_3_3_6(unit:TileUnit) -> None:
  """Set up (3.3.3.3.6)."""
  h = unit.spacing * np.sqrt(3/7)
  t = h / 2
  hexagon = tiling_utils.get_regular_polygon(h, 6)
  tri = tiling_utils.get_regular_polygon(t, 3)
  tri_up = affine.translate(tri, 0, -tri.bounds[3])
  tri_dn = affine.rotate(tri_up, 180, origin = (0, 0))
  wdth, hght = tri.bounds[2] - tri.bounds[0], tri.bounds[3] - tri.bounds[1]
  triangles = [affine.translate(tri_up,        0,  2*hght),
               affine.translate(tri_up,   wdth  ,  2*hght),
               affine.translate(tri_dn,   wdth  ,       0),
               affine.translate(tri_up, 3*wdth/2,    hght),
               affine.translate(tri_dn, 3*wdth/2,   -hght),
               affine.translate(tri_up,   wdth  ,       0),
               affine.translate(tri_dn,   wdth  , -2*hght),
               affine.translate(tri_dn,        0, -2*hght)]
  polys = [hexagon, *triangles]
  unit.setup_vectors((-wdth/2, 3*hght), (2*wdth, 2*hght), (5*wdth/2, -hght))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcdefghi")}, crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_archimedean_3_3_3_4_4(unit:TileUnit) -> None:
  """Set up (3.3.3.4.4)."""
  s = unit.spacing / np.sqrt(1 + sqrt3)
  t = s * sqrt3/2
  square = tiling_utils.get_regular_polygon(s, 4)
  tri = tiling_utils.get_regular_polygon(t, 3)
  tri_up = affine.translate(tri, 0, -tri.bounds[1] + s/2)
  tri_dn = affine.rotate(tri_up, 180, origin = (0, 0))
  polys = [square, tri_up, tri_dn]
  unit.setup_vectors((s/2, s+t), (s, 0), (s/2, -s-t))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abc")}, crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_archimedean_3_3_4_3_4(unit:TileUnit) -> None:
  """Set up (3.3.4.3.4)."""
  s = unit.spacing/2/np.cos(np.pi/12)
  t = s * sqrt3/2
  square = tiling_utils.get_regular_polygon(s, 4)
  tri = tiling_utils.get_regular_polygon(t, 3)
  tri1 = affine.translate(tri, 0, -tri.bounds[3])
  tri2 = affine.rotate(tri1,  180, origin = (0, -t))
  tri3 = affine.rotate(tri1,  150, origin = (0,  0))
  tri4 = affine.rotate(tri1, -150, origin = (0,  0))
  sq1 = affine.translate(square, s/2, -s/2)
  sq1 = affine.rotate(sq1,   30, origin = (0, 0))
  sq2 = affine.rotate(sq1, -150, origin = (0, 0))
  polys = [tri2, tri1, sq2, tri4, tri3, sq1]
  polys = [affine.translate(p, 0, t - s/2) for p in polys]
  unit.setup_vectors((t + s/2,  t + s/2), (t + s/2, -t - s/2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcdef")}, crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_archimedean_3_4_6_4(unit:TileUnit) -> None:
  """Set up (3.4.6.4)."""
  h = unit.spacing / (1 + sqrt3) * sqrt3
  s = h / sqrt3
  t = h / 2
  hexagon = tiling_utils.get_regular_polygon(h, 6)
  square = tiling_utils.get_regular_polygon(s, 4)
  tri = tiling_utils.get_regular_polygon(t, 3)
  sq1 = affine.translate(square, 0, s / 2 + t)
  sq2 = affine.rotate(sq1,  60, origin = (0, 0))
  sq3 = affine.rotate(sq1, -60, origin = (0, 0))
  tri1 = affine.translate(tri, 0, t - tri.bounds[1])
  tri2 = affine.translate(tri, 0, t - tri.bounds[1])
  tri1 = affine.rotate(tri1, 90, origin = (-s / 2, t))
  tri2 = affine.rotate(tri2, -90, origin = (s / 2, t))
  polys = [hexagon, sq2, tri1, sq1, tri2, sq3]
  unit.setup_vectors((0, h + s), (t + 3*s/2,  t + s/2), (t + 3*s/2, -t - s/2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcdef")}, crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_archimedean_3_6_3_6(unit:TileUnit) -> None:
  """Set up (3.6.3.6)."""
  h = unit.spacing * sqrt3/2
  t = h / 2
  hexagon = tiling_utils.get_regular_polygon(h, 6)
  tri = tiling_utils.get_regular_polygon(t, 3)
  tri1 = affine.translate(tri, 0, t - tri.bounds[1])
  tri2 = affine.rotate(tri, 180)
  tri2 = affine.translate(tri2, 0, -t - tri.bounds[3])
  polys = [hexagon, tri1, tri2]
  unit.setup_vectors((t*2/sqrt3,  h), (t*2/sqrt3, -h))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abc")}, crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_archimedean_3_12_12(unit:TileUnit) -> None:
  """Set up (3.12.12)."""
  d = unit.spacing
  t = d * sqrt3/2 / (2 + sqrt3)
  dodecagon = tiling_utils.get_regular_polygon(d, 12)
  tri = tiling_utils.get_regular_polygon(t, 3)
  tri = affine.translate(tri, 0, unit.spacing / 2 - tri.bounds[1])
  tri_1 = affine.rotate(tri,  30, origin = (0, 0))
  tri_2 = affine.rotate(tri, -30, origin = (0, 0))
  polys = [dodecagon, tri_1, tri_2]
  unit.setup_vectors((0, d), (d * sqrt3/2,  d/2), (d * sqrt3/2, -d/2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abc")}, crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_archimedean_4_6_12(unit:TileUnit) -> None:
  """Set up (4.6.12)."""
  d = unit.spacing * (2 + sqrt3) / (3 + sqrt3)
  s = d / (2 + sqrt3)
  h = s * sqrt3
  dodecagon = tiling_utils.get_regular_polygon(d, 12)
  hexagon = tiling_utils.get_regular_polygon(h, 6)
  square = tiling_utils.get_regular_polygon(s, 4)
  sq1 = affine.translate(square, 0, (s + d) / 2)
  sq2 = affine.rotate(sq1,  60, origin = (0, 0))
  sq3 = affine.rotate(sq1, -60, origin = (0, 0))
  hex1 = affine.translate(hexagon, 0, (h + d) / 2)
  hex1 = affine.rotate(hex1,  30, origin = (0, 0))
  hex2 = affine.rotate(hex1, -60, origin = (0, 0))
  polys = [dodecagon, sq2, hex1, sq1, hex2, sq3]
  unit.setup_vectors((                0,   d + s     ),
                     ((d + s) * sqrt3/2,  (d + s) / 2),
                     ((d + s) * sqrt3/2, -(d + s) / 2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcdef")}, crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_archimedean_4_8_8(unit:TileUnit) -> None:
  """Set up (4.8.8)."""
  o = unit.spacing
  s = o / (1 + sqrt2)
  octagon = tiling_utils.get_regular_polygon(o, 8)
  square = tiling_utils.get_regular_polygon(s, 4)
  square = affine.translate(square, 0, (o + s) / 2)
  square = affine.rotate(square, 45, origin = (0, 0))
  polys = [octagon, square]
  unit.setup_vectors((0, o), (o, 0))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("ab")}, crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_hex_colouring(unit:TileUnit) -> str|None:
  """Set up colourings of a regular array of hexagons.

  The supplied unit must have unit.n set.

  Args:
    unit (TileUnit): the TileUnit to setup.

  """
  if any(k not in unit.__dict__ for k in ["n"]):
    return ("""Hex colouring tiling requires n to be supplied.""")
  hexagon = tiling_utils.get_regular_polygon(unit.spacing / np.sqrt(unit.n), 6)
  bounds = hexagon.bounds
  w = bounds[2] - bounds[0]
  h = bounds[3] - bounds[1]
  match unit.n:
    case 2:
      # Pair of point up hexes sideways displaced by width of a hex
      # Vectors are rectangular
      hexagon = affine.rotate(hexagon, 30, origin = (0, 0))
      w, h = h, w
      hexes = [affine.translate(hexagon, dx, 0) for dx in [-w/2, w/2]]
      unit.setup_vectors((w/2, 3*h/4), (2*w, 0))
    case 3:
      # Point up hex at '*' displaced to 3 positions:
      #      2
      #      *
      #    3   1
      hexagon = affine.rotate(hexagon, 30, origin = (0, 0))
      w, h = h, w
      # Copy and translate to alternate corners
      corners = [p for i, p in enumerate(hexagon.exterior.coords)
                 if i in (0, 2, 4)]
      hexes = [affine.translate(hexagon, p[0], p[1]) for p in corners]
      unit.setup_vectors((0  ,  3*h/2), (3*w/2,  3*h/4), (3*w/2, -3*h/4))
    case 4:
      # Point up hex at '*' displaced to 4 positions:
      #      2
      #     3*1
      #      4
      hexagon = affine.rotate(hexagon, 30, origin = (0, 0))
      w, h = h, w
      hex1 = affine.translate(hexagon, unit.spacing / 4, 0)
      hex2 = affine.translate(hexagon, 0, unit.spacing * sqrt3 / 4)
      hex3 = affine.translate(hexagon, -unit.spacing / 4, 0)
      hex4 = affine.translate(hexagon, 0, -unit.spacing * sqrt3 / 4)
      hexes = [hex1, hex2, hex3, hex4]
      unit.setup_vectors((w,  3*h/2), (w, -3*h/2))
    case 4:
      # Point up hex at '*' displaced to 4 positions:
      #      2
      #     3*1
      #      4
      hexagon = affine.rotate(hexagon, 30, origin = (0, 0))
      w, h = h, w
      hex1 = affine.translate(hexagon, unit.spacing / 4, 0)
      hex2 = affine.translate(hexagon, 0, unit.spacing * sqrt3 / 4)
      hex3 = affine.translate(hexagon, -unit.spacing / 4, 0)
      hex4 = affine.translate(hexagon, 0, -unit.spacing * sqrt3 / 4)
      hexes = [hex1, hex2, hex3, hex4]
      unit.setup_vectors((w,  3*h/2), (w, -3*h/2))
    case 5:
      hexagon = affine.rotate(hexagon, 30, origin = (0, 0))
      w, h = h, w
      tr = [(-w,  h/4), (0,  h/4), ( w, h/4), (-w/2, -h/2), (w/2, -h/2)]
      hexes = [affine.translate(hexagon, v[0], v[1]) for v in tr]
      unit.setup_vectors((0,  3*h/2), (5*w/2,  3*h/4), (5*w/2, -3*h/4))
    case 6:
      hexagon = affine.rotate(hexagon, 30, origin = (0, 0))
      w, h = h, w
      tr = [(-5*w/4,  3*h/8), (-w/4,  3*h/8), (3*w/4,  3*h/8),
            (-3*w/4, -3*h/8), ( w/4, -3*h/8), (5*w/4, -3*h/8)]
      hexes = [affine.translate(hexagon, v[0], v[1]) for v in tr]
      unit.setup_vectors((0,  3*h/2), (3*w, 0))
    case 7:  # the 'H3' tile
      # Make a hexagon and displace in the direction of its
      # own 6 corners, scaled as needed
      base_hex = tiling_utils.get_regular_polygon(unit.spacing, 6)
      bounds = base_hex.bounds
      w = bounds[2] - bounds[0]
      h = bounds[3] - bounds[1]
      rotation = np.degrees(np.arctan(1 / 3 / sqrt3))
      corners = list(hexagon.exterior.coords)[:-1]
      hexagon = affine.rotate(hexagon, 30)
      hexes = [hexagon] + [affine.translate(
        hexagon, x * sqrt3, y * sqrt3) for x, y in corners]
      hexes = [affine.rotate(h, rotation, origin = (0, 0))
              for h in hexes]
      unit.setup_vectors((0,  h), (3*w/4,  h/2), (3*w/4, -h/2))
    case 8:
      #   1 2
      #  3 4 5
      #   6 7
      #    8
      hexagon = affine.rotate(hexagon, 30, (0, 0))
      w, h = h, w
      offsets = [ (-w/2,  3*h/4), (w/2,  3*h/4),
                  (-w, 0),    (0, 0),   ( w, 0),
                  (-w/2, -3*h/4), (w/2, -3*h/4),
                          (0, -3*h/2)]
      hexes = [affine.translate(hexagon, dx, dy) for dx, dy in offsets]
      unit.setup_vectors((0, 3*h), (2*w, 3*h/2), (2*w, -3*h/2))
    case 9:
      #    1
      #   2 3
      #  4 5 6
      #   7 8
      #    9
      hexagon = affine.rotate(hexagon, 30, (0, 0))
      w, h = h, w
      offsets = [         (0,  3*h/2),
                  (-w/2,  3*h/4), (w/2,  3*h/4),
                  (-w, 0),    (0, 0),   ( w, 0),
                  (-w/2, -3*h/4), (w/2, -3*h/4),
                          (0, -3*h/2)]
      hexes = [affine.translate(hexagon, dx, dy) for dx, dy in offsets]
      unit.setup_vectors((3*w/2, 9*h/4), (3*w/2, -9*h/4))
    case 10:
      #   1 2 3
      #  4 5 6 7
      #   8 9 10
      hexagon = affine.rotate(hexagon, 30, (0, 0))
      w, h = h, w
      offsets = [ (-w, 3*h/4), (0, 3*h/4), (w, 3*h/4),
                 (-3*w/2, 0), (-w/2, 0), (w/2, 0), (3*w/2, 0),
                 (-w, -3*h/4), (0, -3*h/4), (w, -3*h/4)]
      hexes = [affine.translate(hexagon, dx, dy) for dx, dy in offsets]
      unit.setup_vectors((w/2, 9*h/4), (7*w/2, 3*h/4), (3*w, -3*h/2))
    case 11:
      #   1  2  3
      #  4  5  6  7  8
      #      9  10 11
      hexagon = affine.rotate(hexagon, 30, (0, 0))
      w, h = h, w
      offsets = [(-3*w/2, 3*h/4), (-w/2, 3*h/4), (w/2, 3*h/4),
                 (-2*w, 0), (-w, 0), (0, 0), (w, 0), (2*w, 0),
                 (-w/2, -3*h/4), (w/2, -3*h/4), (3*w/2, -3*h/4)]
      hexes = [affine.translate(hexagon, dx, dy) for dx, dy in offsets]
      unit.setup_vectors((-w/2, 9*h/4), (7*w/2, 3*h/4), (4*w, -3*h/2))
    case 16:
      #       1
      #     2   3
      #   4   5   6
      # 7   8   9  10
      #  11  12  13
      #    14  15
      #      16
      hexagon = affine.rotate(hexagon, 30, (0, 0))
      w, h = h, w
      offsets = [         (0,  9*h/4),
                  (-w/2,  3*h/2), (w/2,  3*h/2),
              (-w,  3*h/4), (0,  3*h/4),  ( w,  3*h/4),
          (-3*w/2, 0), (-w/2, 0), ( w/2, 0), ( 3*w/2, 0),
              (-w, -3*h/4), (0, -3*h/4),  ( w, -3*h/4),
                  (-w/2, -3*h/2), (w/2, -3*h/2),
                          (0, -9*h/4)]
      hexes = [affine.translate(hexagon, dx, dy) for dx, dy in offsets]
      unit.setup_vectors((2*w, 3*h), (2*w, -3*h))
    case _:
      return (f"""{unit.n}-colouring of hexes is not supported.
              Try a number between 2 and 9, or 16.""")
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:unit.n]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(hexes))
  return None


def _setup_square_colouring(unit:TileUnit) -> str|None:
  """Set up colourings of a regular array of squares.

  The supplied unit must have unit.n set.

  Args:
    unit (TileUnit):  the TileUnit to setup.

  """
  if any(k not in unit.__dict__ for k in ["n"]):
    return "Square colouring tiling requires n to be supplied."
  sq = tiling_utils.get_regular_polygon(unit.spacing / np.sqrt(unit.n), 4)
  s = sq.bounds[2] - sq.bounds[0]
  match unit.n:
    case 2:
      # Copy and translate square
      tr = [(-s/2, 0), (s/2, 0)]
      squares = [affine.translate(sq, v[0], v[1]) for v in tr]
      unit.setup_vectors((0, s), (2*s, 0))
    case 3:
      # Copy and translate square
      tr = [(-s/2, -s/2), (s/2, -s/2), (-s/2, s/2)]
      squares = [affine.translate(sq, v[0], v[1]) for v in tr]
      squares = [affine.rotate(sq, 45, (0, 0)) for sq in squares]
      x = s * sqrt2
      unit.setup_vectors((0,  x), (3*x/2,  x/2), (3*x/2, -x/2))
    case 4:
      # Copy and translate square
      tr = [(-s/2, -s/2), (-s/2,  s/2),
            ( s/2,  s/2), ( s/2, -s/2)]
      squares = [affine.translate(sq, v[0], v[1]) for v in tr]
      unit.setup_vectors((0, unit.spacing), (unit.spacing, 0))
    case 5:
      # Copy and translate square
      tr = [( 0, 0),
            ( s, 0), (0,  s),
            (-s, 0), (0, -s)]
      squares = [affine.translate(sq, v[0], v[1]) for v in tr]
      rotation = np.degrees(np.arctan2(1, 2))
      squares = [affine.rotate(sq, rotation, origin = (0, 0))
                 for sq in squares]
      unit.setup_vectors((0, unit.spacing), (unit.spacing, 0))
    case 6:
      # Copy and translate square
      tr = [(-s, -s), (0, -s),
            (-s,  0), (0,  0), ( s,  0),
                      (0,  s)]
      squares = [affine.translate(sq, v[0], v[1]) for v in tr]
      squares = [affine.rotate(sq, -45, origin = (0, 0)) for sq in squares]
      x = s * sqrt2
      unit.setup_vectors(( x/2,  3*x/2), (2*x,  0), (3*x/2, -3*x/2))
    case 7:
      # Copy and translate square
      tr = [(-s, -s), (0, -s),
            (-s,  0), (0,  0), ( s,  0),
                      (0,  s), ( s,  s)]
      squares = [affine.translate(sq, v[0], v[1]) for v in tr]
      squares = [affine.rotate(sq, -45, (0, 0)) for sq in squares]
      x = s * sqrt2
      unit.setup_vectors(( x/2,  3*x/2), (5*x/2,  x/2), (2*x, -x))
    case 8:
      # Copy and translate square
      tr = [(-s, -s), (0, -s), ( s, -s),
            (-s,  0), (0,  0), ( s,  0),
            (-s,  s), (0,  s)]
      squares = [affine.translate(sq, v[0], v[1]) for v in tr]
      squares = [affine.rotate(sq, 45, origin = (0, 0)) for sq in squares]
      x  = s * sqrt2
      unit.setup_vectors((0, 2*x), (2*x,  x), (2*x,  -x))
    case 9:
      # Copy and translate square
      tr = [(-s, -s), (0, -s), ( s, -s),
            (-s,  0), (0,  0), ( s,  0),
            (-s,  s), (0,  s), ( s,  s)]
      squares = [affine.translate(sq, v[0], v[1]) for v in tr]
      unit.setup_vectors((0, unit.spacing), (unit.spacing, 0))
    case 16:
      # Copy and translate square
      tr = [(-3*s/2, -3*s/2), (-s/2, -3*s/2), ( s/2, -3*s/2), ( 3*s/2, -3*s/2),
            (-3*s/2,   -s/2), (-s/2,   -s/2), ( s/2,   -s/2), ( 3*s/2,   -s/2),
            (-3*s/2,    s/2), (-s/2,    s/2), ( s/2,    s/2), ( 3*s/2,    s/2),
            (-3*s/2,  3*s/2), (-s/2,  3*s/2), ( s/2,  3*s/2), ( 3*s/2,  3*s/2)]
      squares = [affine.translate(sq, v[0], v[1]) for v in tr]
      unit.setup_vectors((0, unit.spacing), (unit.spacing, 0))
    case _:
      return (f"""{unit.n}-colouring of squares is not supported.
              Try a number between 2 and 9, or 16.""")

  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:unit.n]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(squares))
  return None


def _get_corners_and_mid_points(
      shape:geom.Polygon,
      offset:float = 0.5,
    ) -> list[geom.Point]:
  """Return corners and midpoints of polygon.

  Args:
    shape (geom.Polygon): polygon for which corners and midpoints are requested.
    offset (float): how far along sides midpoints should be.

  Returns:
    list[geom.Point]: list of points in order (i.e corners and midpoints are
      interleaved).

  """
  corners = [geom.Point(p) for p in shape.exterior.coords][:-1]
  midpoints = [ls.interpolate(offset, normalized = True)
               for ls in tiling_utils.get_sides(shape)]
  return list(itertools.chain(*[
                [a, b] for a, b in zip(corners, midpoints, strict = True)]))


def _setup_square_dissection(unit:TileUnit) -> str|None:
  """Set up dissection of a square."""
  if any(k not in unit.__dict__ for k in ["n", "offset", "offset_angle"]):
    return ("""Square dissection tiling requires n, offset, and offset_angle
            to be be supplied.""")
  outer = tiling_utils.get_regular_polygon(unit.spacing, 4)
  inner = affine.scale(outer, 1 / np.sqrt(unit.n), 1 / np.sqrt(unit.n))
  inner = affine.rotate(inner, unit.offset_angle)
  outer_pts = _get_corners_and_mid_points(outer)
  inner_pts = _get_corners_and_mid_points(inner)
  outer_pts = outer_pts + outer_pts[:1]
  inner_pts = inner_pts + inner_pts[:1]
  inner_pts = inner_pts[::-1]
  match unit.n:
    case 3:
      polys = [inner]
      if unit.offset == 1:
        polys.extend([
          geom.Polygon(outer_pts[1:3] + outer_pts[4:6] + inner_pts[4:9:2]),
          geom.Polygon(outer_pts[5:7] + outer_pts[0:2] + inner_pts[0:5:2])])
      else:
        polys.extend([
          geom.Polygon(outer_pts[0:5:2] + inner_pts[4:9:2]),
          geom.Polygon(outer_pts[4:9:2] + inner_pts[0:5:2])])
    case 5:
      polys = [inner]
      if unit.offset == 1:
        polys.extend([
          geom.Polygon(outer_pts[1:4] + inner_pts[6:9:2]),
          geom.Polygon(outer_pts[3:6] + inner_pts[4:7:2]),
          geom.Polygon(outer_pts[5:8] + inner_pts[2:5:2]),
          geom.Polygon(outer_pts[7:9] + outer_pts[1:2] + inner_pts[0:3:2])])
      else:
        polys.extend([
          geom.Polygon(outer_pts[0:3:2] + inner_pts[6:9:2]),
          geom.Polygon(outer_pts[2:5:2] + inner_pts[4:7:2]),
          geom.Polygon(outer_pts[4:7:2] + inner_pts[2:5:2]),
          geom.Polygon(outer_pts[6:9:2] + inner_pts[0:3:2])])
    case 9:
      polys = [inner]
      if unit.offset == 1:
        polys.extend([
          geom.Polygon(outer_pts[1:3] + inner_pts[7:9]),
          geom.Polygon(outer_pts[2:4] + inner_pts[6:8]),
          geom.Polygon(outer_pts[3:5] + inner_pts[5:7]),
          geom.Polygon(outer_pts[4:6] + inner_pts[3:6]),
          geom.Polygon(outer_pts[5:7] + inner_pts[3:5]),
          geom.Polygon(outer_pts[6:8] + inner_pts[2:4]),
          geom.Polygon(outer_pts[7:9] + inner_pts[1:3]),
          geom.Polygon(outer_pts[0:2] + inner_pts[0:2])])
      else:
        polys.extend([
          geom.Polygon(outer_pts[0:2] + inner_pts[7:9]),
          geom.Polygon(outer_pts[1:3] + inner_pts[6:8]),
          geom.Polygon(outer_pts[2:4] + inner_pts[5:7]),
          geom.Polygon(outer_pts[3:5] + inner_pts[4:6]),
          geom.Polygon(outer_pts[4:6] + inner_pts[3:5]),
          geom.Polygon(outer_pts[5:7] + inner_pts[2:4]),
          geom.Polygon(outer_pts[6:8] + inner_pts[1:3]),
          geom.Polygon(outer_pts[7:9] + inner_pts[0:2])])
    case _:
      unit.base_shape = TileShape.RECTANGLE
      return f"{unit.n} dissection of square not available"
  unit.setup_vectors((0, unit.spacing), (unit.spacing, 0))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:unit.n]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))
  return None


def _setup_hex_dissection(unit:TileUnit):
  """Set up dissection of a hexagon."""
  if any(k not in unit.__dict__ for k in ["n", "offset", "offset_angle"]):
    return ("""Hex dissection tiling requires n, offset, and offset_angle
            to be be supplied.""")
  outer = tiling_utils.get_regular_polygon(unit.spacing, 6)
  if unit.n == 9:
    inner = affine.scale(outer, 1/np.sqrt(3), 1/np.sqrt(3))
  else:
    inner = affine.scale(outer, 1 / np.sqrt(unit.n), 1 / np.sqrt(unit.n))
  inner = affine.rotate(inner, unit.offset_angle)
  outer_pts = _get_corners_and_mid_points(outer)
  inner_pts = _get_corners_and_mid_points(inner)
  outer_pts = outer_pts + outer_pts[:1]
  inner_pts = inner_pts + inner_pts[:1]
  inner_pts = inner_pts[::-1]
  match unit.n:
    case 3:
      polys = [inner]
      if unit.offset == 1:
        polys.extend([
          geom.Polygon(outer_pts[1:3] + outer_pts[2:7:2] +
                       outer_pts[7:8] + inner_pts[6::2]),
          geom.Polygon(outer_pts[7:9] + outer_pts[8::2] +
                       outer_pts[0:2] + inner_pts[0:7:2])])
      else:
        polys.extend([
          geom.Polygon(outer_pts[0:7:2] + inner_pts[6::2]),
          geom.Polygon(outer_pts[6::2]  + inner_pts[0:7:2])])
    case 4:
      polys = [inner]
      if unit.offset == 1:
        polys.extend([
          geom.Polygon(outer_pts[1:3]  + outer_pts[2:5:2] +
                       outer_pts[4:6]  + inner_pts[8::2]),
          geom.Polygon(outer_pts[5:7]  + outer_pts[6:11:2] +
                       outer_pts[8:10] + inner_pts[4:9:2]),
          geom.Polygon(outer_pts[9:11] + outer_pts[10::2] +
                       outer_pts[0:2]  + inner_pts[0:5:2])])
      else:
        polys.extend([
          geom.Polygon(outer_pts[0:5:2] + inner_pts[8::2]),
          geom.Polygon(outer_pts[4:9:2] + inner_pts[4:9:2]),
          geom.Polygon(outer_pts[8::2]  + inner_pts[0:5:2])])
    case 7:
      polys = [inner]
      if unit.offset == 1:
        polys.extend([
          geom.Polygon(outer_pts[1:3]  + outer_pts[2:4]   + inner_pts[10::2]),
          geom.Polygon(outer_pts[3:5]  + outer_pts[4:6]   + inner_pts[8:11:2]),
          geom.Polygon(outer_pts[5:7]  + outer_pts[6:8]   + inner_pts[6:9:2]),
          geom.Polygon(outer_pts[7:9]  + outer_pts[8:10]  + inner_pts[4:7:2]),
          geom.Polygon(outer_pts[9:11] + outer_pts[10:12] + inner_pts[2:5:2]),
          geom.Polygon(outer_pts[11:]  + outer_pts[:2]    + inner_pts[0:3:2])])
      else:
        polys.extend([
          geom.Polygon(outer_pts[0:3:2]  + inner_pts[10::2]),
          geom.Polygon(outer_pts[2:5:2]  + inner_pts[8:11:2]),
          geom.Polygon(outer_pts[4:7:2]  + inner_pts[6:9:2]),
          geom.Polygon(outer_pts[6:9:2]  + inner_pts[4:7:2]),
          geom.Polygon(outer_pts[8:11:2] + inner_pts[2:5:2]),
          geom.Polygon(outer_pts[10::2]  + inner_pts[0:3:2])])
    case 9:
      if unit.offset == 1:
        polys = [
          geom.Polygon(inner_pts[1:3]  + inner_pts[2:5:2] +
                       inner_pts[4:6]  + [(0, 0)]),
          geom.Polygon(inner_pts[5:7]  + inner_pts[6:9:2] +
                       inner_pts[8:10] + [(0, 0)]),
          geom.Polygon(inner_pts[9:11] + inner_pts[10::2] +
                       inner_pts[0:2]  + [(0, 0)]),
          geom.Polygon(outer_pts[1:3]  + outer_pts[2:4]   + inner_pts[10::2]),
          geom.Polygon(outer_pts[3:5]  + outer_pts[4:6]   + inner_pts[8:11:2]),
          geom.Polygon(outer_pts[5:7]  + outer_pts[6:8]   + inner_pts[6:9:2]),
          geom.Polygon(outer_pts[7:9]  + outer_pts[8:10]  + inner_pts[4:7:2]),
          geom.Polygon(outer_pts[9:11] + outer_pts[10:12] + inner_pts[2:5:2]),
          geom.Polygon(outer_pts[11:]  + outer_pts[:2]    + inner_pts[0:3:2])]
      else:
        polys = [
          geom.Polygon(inner_pts[0:5:2]  + [(0, 0)]),
          geom.Polygon(inner_pts[4:9:2]  + [(0, 0)]),
          geom.Polygon(inner_pts[8::2]   + [(0, 0)]),
          geom.Polygon(outer_pts[0:3:2]  + inner_pts[10::2]),
          geom.Polygon(outer_pts[2:5:2]  + inner_pts[8:11:2]),
          geom.Polygon(outer_pts[4:7:2]  + inner_pts[6:9:2]),
          geom.Polygon(outer_pts[6:9:2]  + inner_pts[4:7:2]),
          geom.Polygon(outer_pts[8:11:2] + inner_pts[2:5:2]),
          geom.Polygon(outer_pts[10::2]  + inner_pts[0:3:2])]
    case _:
      unit.base_shape = TileShape.HEXAGON
      return f"{unit.n} dissection of hexagon not available"
  unit.setup_vectors((0, unit.spacing),
                     (unit.spacing * sqrt3/2, unit.spacing / 2),
                     (unit.spacing * sqrt3/2, -unit.spacing / 2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:unit.n]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))
  return None


def _setup_star_polygon_1(unit:TileUnit) -> str|None:
  """Set up a star-polygon tiling."""
  if any(k not in unit.__dict__ for k in ["code", "point_angle"]):
    return ("""Star tiling 1 requires code and point_angle
            to be be supplied.""")
  match unit.code:
    case "33":
      _setup_star_polys_33(unit)
    case "36":
      _setup_star_polys_36(unit)
    case "44":
      _setup_star_polys_44(unit)
    case "63":
      _setup_star_polys_63(unit)
    case _:
      return "Invalid code supplied for star tiling 1"
  # star polygons cause issues for _setup_regularised_prototile
  # due to negative buffering required in unioning, so each of
  # these calculates a unit.union polygon and we assign here
  unit.regularised_prototile = gpd.GeoDataFrame(
    geometry = gpd.GeoSeries([unit.union]), crs = unit.crs)
  return None

def _setup_star_polys_33(unit:TileUnit) -> None:
  """Set up polygon by 3-star and triangles."""
  star = _get_star(3, unit.spacing / 2, unit.point_angle)
  s_sides = tiling_utils.get_sides(star)
  tri1 = _get_regular_polygon_from_one_side(s_sides[0], 3)
  tri2 = _get_regular_polygon_from_one_side(s_sides[5], 3)
  unit.union = geom.Polygon(tri1.exterior.coords[2:0:-1] +
                       star.exterior.coords[2:6] +
                       tri2.exterior.coords[2:0:-1])
  tips = list(star.exterior.coords)[0::2]
  dents = list(star.exterior.coords)[1::2]
  v1 = geom.Point(dents[0][0] - tips[2][0], dents[0][1] - tips[2][1])
  v2 = affine.rotate(v1, -60, (0, 0))
  v3 = affine.rotate(v2, -60, (0, 0))
  unit.setup_vectors((v1.x, v1.y), (v2.x, v2.y), (v3.x, v3.y))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abc")},
    crs = unit.crs,
    geometry = gpd.GeoSeries([star, tri1, tri2]))


def _setup_star_polys_36(unit:TileUnit) -> None:
  """Set up tiling by 3-star and hexagons."""
  star1 = _get_star(3, unit.spacing / 3, unit.point_angle)
  hexagon = _get_regular_polygon_from_one_side(
    tiling_utils.get_sides(star1)[0], 6)
  c = hexagon.centroid
  star1 = affine.translate(star1, -c.x, -c.y)
  hexagon = affine.translate(hexagon, -c.x, -c.y)
  star2 = affine.rotate(star1, 180, (0, 0))
  unit.union = geom.Polygon(
    hexagon.exterior.coords[5:3:-1] + star2.exterior.coords[1:7] +
    hexagon.exterior.coords[2:0:-1] + star1.exterior.coords[1:7])
  tips1 = list(star1.exterior.coords)[0::2]
  dents2 = list(star2.exterior.coords)[1::2]
  v1 = geom.Point(dents2[2][0] - tips1[0][0], dents2[2][1] - tips1[0][1])
  v2 = affine.rotate(v1, -60, (0, 0))
  unit.setup_vectors((v1.x, v1.y), (v2.x, v2.y))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abc")},
    crs = unit.crs,
    geometry = gpd.GeoSeries([star1, star2, hexagon]))


def _setup_star_polys_44(unit:TileUnit) -> None:
  """Set up tiling by 4-stars and squares."""
  star = _get_star(4, unit.spacing / 2, unit.point_angle)
  square = _get_regular_polygon_from_one_side(
    tiling_utils.get_sides(star)[0], 4)
  unit.union = geom.Polygon(
    square.exterior.coords[3:1:-1] + star.exterior.coords[1:])
  tips = list(star.exterior.coords)[0::2]
  dents = list(star.exterior.coords)[1::2]
  v1 = geom.Point(tips[0][0] - dents[1][0], tips[0][1] - dents[1][1])
  v2 = affine.rotate(v1, 90, (0, 0))
  unit.setup_vectors((v1.x, v1.y), (v2.x, v2.y))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("ab")},
    crs = unit.crs,
    geometry = gpd.GeoSeries([star, square]))


def _setup_star_polys_63(unit:TileUnit) -> None:
  """Set up tiling by 6-star and triangles."""
  star = _get_star(6, unit.spacing / 2, unit.point_angle)
  s_sides = tiling_utils.get_sides(star)
  tri1 = _get_regular_polygon_from_one_side(s_sides[0], 3)
  tri2 = _get_regular_polygon_from_one_side(s_sides[6], 3)
  unit.union = geom.Polygon(
    tri1.exterior.coords[2:0:-1] + star.exterior.coords[2:7] +
    tri2.exterior.coords[2:0:-1] + star.exterior.coords[8:])
  tips = list(star.exterior.coords)[0::2]
  dents = list(star.exterior.coords)[1::2]
  v1 = geom.Point(dents[0][0] - tips[4][0], dents[0][1] - tips[4][1])
  v2 = affine.rotate(v1, -60, (0, 0))
  v3 = affine.rotate(v2, -60, (0, 0))
  unit.setup_vectors((v1.x, v1.y), (v2.x, v2.y), (v3.x, v3.y))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abc")},
    crs = unit.crs,
    geometry = gpd.GeoSeries([star, tri1, tri2]))


def _get_regular_polygon_from_one_side(
    side:geom.LineString,
    n:int,
  ) -> geom.Polygon:
  """Return a regular polygon given one side.

  Args:
    side (geom.LineString): one side of the requested polygon.
    n (int): number of sides of the polygon.

  Returns:
    geom.Polygon: the polygon.

  """
  points = list(side.coords)
  turn_angle = np.pi - 2 * np.pi / n
  for _ in range(1, n):
    points.append(
      affine.rotate(geom.Point(points[-2]),
                    turn_angle, origin = points[-1],
                    use_radians = True))
  return geom.Polygon(points[:-1])


def _get_star(
    n:int,
    point_length:float,
    point_angle:float,
  ) -> geom.Polygon:
  """Return a star polygon.

  Args:
    n (int): the number of points.
    point_length(float): the length of start points.
    point_angle(float): angle at the tip of each point.

  Returns:
    geom.Polygon: the star polygon.

  """
  pt_angle = np.radians(point_angle)
  hp = point_length * (np.cos(pt_angle / 2) + np.sin(pt_angle / 2) /
                       np.tan(np.pi / n))
  hd = point_length * np.sin(pt_angle / 2) / np.sin(np.pi / n)
  interior_angles = np.linspace(0, 2 * np.pi, 2 * n, endpoint = False)
  tips = [geom.Point(hp * np.cos(a), hp * np.sin(a))
          for a in interior_angles[::2]]
  dents = [geom.Point(hd * np.cos(a), hd * np.sin(a))
             for a in interior_angles[1::2]]
  return geom.Polygon(list(itertools.chain(*zip(tips, dents, strict = True))))


def _setup_star_polygon_2(unit:TileUnit):
  """Set up a star-polygon tiling with no tiling point angle option."""
  if any(k not in unit.__dict__ for k in ["code"]):
    return "Star tiling 2 requires code to be be supplied."
  match unit.code:
    case "45":
      _setup_star_polys_45(unit)
    case "464":
      _setup_star_polys_464(unit)
    case "466":
      _setup_star_polys_466(unit)
    case "64":
      _setup_star_polys_64(unit)
    case "66":
      _setup_star_polys_66(unit)
    case "663":
      _setup_star_polys_663(unit)
    case "844":
      _setup_star_polys_844(unit)
    case _:
      return "Invalid code supplied for star tiling 2"
  # star polygons cause issues for _setup_regularised_prototile
  # due to negative buffering required in unioning, so each of
  # these calculates a unit.union polygon and we assign here
  unit.regularised_prototile = gpd.GeoDataFrame(
    geometry = gpd.GeoSeries([unit.union]), crs = unit.crs)
  return None


def _setup_star_polys_45(unit:TileUnit) -> None:
  """Set up tiling by regular pentagon and 4-stars."""
  pents = []
  pent = tiling_utils.get_regular_polygon(unit.spacing / 2, 5)
  pents.append(affine.translate(pent, 0, -pent.bounds[1]))
  pents.append(
    affine.rotate(pents[0], -126, tiling_utils.get_corners(pents[0])[0]))
  pents.extend([affine.rotate(p, 180, (0, 0)) for p in pents])
  star_diag1 = geom.LineString([
    tiling_utils.get_corners(pents[0])[2],
    affine.rotate(tiling_utils.get_corners(pents[0])[2], -90,
                  tiling_utils.get_corners(pents[0])[0])])
  star_diag2 = affine.rotate(star_diag1, 90)
  star_centre = star_diag1.intersection(star_diag2)
  star_pts = [affine.rotate(
    tiling_utils.get_corners(pents[0])[0], a, star_centre)
    for a in range(0, 360, 90)]
  star_dents = [
    affine.rotate(tiling_utils.get_corners(pents[0])[1], a, star_centre)
    for a in range(0, 360, 90)]
  star1 = geom.Polygon(
    itertools.chain(*zip(star_dents, star_pts, strict = True)))
  star2 = affine.rotate(star1, -126, tiling_utils.get_corners(pents[0])[2])
  unit.union = geom.Polygon(
    star1.exterior.coords[4:8] + star2.exterior.coords[:4] +
    pents[3].exterior.coords[3:] + pents[2].exterior.coords[1:5] +
    pents[1].exterior.coords[1:4])
  ppts = list(pents[0].exterior.coords)[:-1]
  spts = list(star1.exterior.coords)[:-1]
  v1 = geom.Point(spts[5][0] - ppts[-1][0], spts[5][0] - ppts[-1][0])
  v2 = affine.rotate(v1, -90, (0, 0))
  unit.setup_vectors((v1.x, v1.y), (v2.x, v2.y))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcdef")},
    geometry = gpd.GeoSeries([*pents, star1, star2]),
    crs = unit.crs)


def _setup_star_polys_464(unit:TileUnit) -> None:
  """Set up 4-star and hexagon and square tiling."""
  star = _get_star(4, unit.spacing / 3, 30)
  hex1 = _get_regular_polygon_from_one_side(tiling_utils.get_sides(star)[0], 6)
  hex2 = _get_regular_polygon_from_one_side(tiling_utils.get_sides(star)[2], 6)
  square = _get_regular_polygon_from_one_side(
    tiling_utils.get_sides(hex2)[5].reverse(), 4)
  unit.union = geom.Polygon(
    hex1.exterior.coords[5:2:-1] + square.exterior.coords[2:1:-1] +
    hex2.exterior.coords[5:1:-1] + star.exterior.coords[5:])
  p0 = tiling_utils.get_corners(star)[6]
  p1 = tiling_utils.get_corners(hex2)[3]
  v1 = geom.Point(p1.x - p0.x, p1.y - p0.y)
  v2 = affine.rotate(v1, -90, (0, 0))
  unit.setup_vectors((v1.x, v1.y), (v2.x, v2.y))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcd")},
    geometry = gpd.GeoSeries([star, hex1, hex2, square]),
    crs = unit.crs)


def _setup_star_polys_466(unit:TileUnit) -> None:
  """Set up 4-star and two hexagon tiling."""
  hex1 = tiling_utils.get_regular_polygon(
    unit.spacing / np.sqrt(1 + sqrt3 / 2), 6)
  hex2 = affine.scale(hex1, 0.5, 0.5)
  hex1 = affine.rotate(hex1, 30)
  w1 = hex1.bounds[2] - hex1.bounds[0]
  w2 = hex2.bounds[2] - hex2.bounds[0]
  hex2 = affine.translate(hex2, w1/2 + w2/2)
  hexes = [hex1] + [affine.rotate(hex2, a, (0, 0)) for a in [60, 120]]
  h1_pts = list(hexes[1].exterior.coords)
  h2_pts = list(hexes[2].exterior.coords)
  long_diag_1 = geom.LineString([h1_pts[4], h2_pts[0]])
  long_diag_2 = affine.rotate(long_diag_1, 90)
  short_diag_1 = geom.LineString([h1_pts[3], h2_pts[5]])
  short_diag_2 = affine.rotate(short_diag_1, 90)
  star = geom.Polygon([long_diag_1.coords[0], short_diag_1.coords[0],
                       long_diag_2.coords[0], short_diag_2.coords[0],
                       long_diag_1.coords[1], short_diag_1.coords[1],
                       long_diag_2.coords[1], short_diag_2.coords[1]])
  stars = [affine.rotate(star, a, (0, 0)) for a in [-60, 0, 60]]
  unit.union = geom.Polygon(
    hexes[0].exterior.coords[4:6] + hexes[0].exterior.coords[0:1] +
    stars[0].exterior.coords[7:2:-1] + hexes[1].exterior.coords[2:4] +
    stars[1].exterior.coords[4:5] + hexes[2].exterior.coords[2:4] +
    stars[2].exterior.coords[4:0:-1])
  p0 = tiling_utils.get_corners(hexes[0])[5]
  p1 = tiling_utils.get_corners(stars[1])[3]
  v1 = geom.Point(p1.x - p0.x, p1.y - p0.y)
  v2, v3 = tuple([affine.rotate(v1, a, (0, 0)) for a in [-60, -120]])
  unit.setup_vectors((v1.x, v1.y), (v2.x, v2.y), (v3.x, v3.y))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcdef")},
    geometry = gpd.GeoSeries(hexes + stars),
    crs = unit.crs)


def _setup_star_polys_64(unit:TileUnit) -> None:
  """Set up 6-star and square tiling."""
  star = _get_star(6, unit.spacing / 2, 30)
  square = _get_regular_polygon_from_one_side(
    tiling_utils.get_sides(star)[0], 4)
  squares = [affine.rotate(square, a, (0, 0)) for a in range(0, 360, 60)]
  unit.union = geom.Polygon(
    squares[0].exterior.coords[3:1:-1] + squares[1].exterior.coords[3:1:-1] +
    squares[2].exterior.coords[3:1:-1] + star.exterior.coords[6:])
  p1 = tiling_utils.get_corners(star)[4]
  p0 = tiling_utils.get_corners(star)[8]
  v1 = geom.Point(p1.x - p0.x, p1.y - p0.y)
  v2, v3 = tuple([affine.rotate(v1, a, (0, 0)) for a in [-60, -120]])
  unit.setup_vectors((v1.x, v1.y), (v2.x, v2.y), (v3.x, v3.y))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcd")},
    geometry = gpd.GeoSeries([star, *squares]),
    crs = unit.crs)


def _setup_star_polys_66(unit:TileUnit) -> None:
  """Set up 6-star and hexagon tiling."""
  star = _get_star(6, unit.spacing / 4, 60)
  hex1 = _get_regular_polygon_from_one_side(tiling_utils.get_sides(star)[0], 6)
  hex2 = _get_regular_polygon_from_one_side(tiling_utils.get_sides(star)[6], 6)
  polys = [affine.rotate(p, -30, (0, 0)) for p in [star, hex1, hex2]]
  unit.union = geom.Polygon(
    polys[1].exterior.coords[5:1:-1] + polys[0].exterior.coords[3:7] +
    polys[2].exterior.coords[5:1:-1] + polys[0].exterior.coords[8:])
  vs = tiling_utils.get_corners(polys[0])[0:12:2]
  p0, p1 = vs[0], vs[3]
  v1 = geom.Point(p1.x - p0.x, p1.y - p0.y)
  v2 = affine.rotate(v1, -120, (0, 0))
  unit.setup_vectors((v1.x, v1.y), (v2.x, v2.y))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abc")},
    geometry = gpd.GeoSeries(polys),
    crs = unit.crs)


def _setup_star_polys_663(unit:TileUnit) -> None:
  """Set up 6-star, hexagon, and triangle tiling."""
  star = _get_star(6, unit.spacing / 4, 60)
  hexagon = _get_regular_polygon_from_one_side(
    tiling_utils.get_sides(star)[0], 6)
  hexes = [affine.rotate(hexagon, a, (0, 0))
           for a in range(180, 360, 60)]
  tri = _get_regular_polygon_from_one_side(
    tiling_utils.get_sides(hexes[0])[2].reverse(), 3)
  triangles = [affine.rotate(tri, a, (0, 0)) for a in [0, 60]]
  unit.union = geom.Polygon(
    star.exterior.coords[0:6] + hexes[0].exterior.coords[-1:-5:-1] +
    hexes[1].exterior.coords[5:2:-1] + hexes[2].exterior.coords[5:2:-1])
  vs = tiling_utils.get_corners(star)[1:12:2]
  p0, p1 = vs[0], vs[3]
  v1 = geom.Point(2 * (p1.x - p0.x), 2 * (p1.y - p0.y))
  v2, v3 = tuple([affine.rotate(v1, a, (0, 0)) for a in [-60, -120]])
  unit.setup_vectors((v1.x, v1.y), (v2.x, v2.y), (v3.x, v3.y))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcdef")},
    geometry = gpd.GeoSeries([star, *hexes, *triangles]),
    crs = unit.crs)


def _setup_star_polys_844(unit:TileUnit) -> None:
  """Set up 8-star and two square tiling."""
  star = _get_star(8, unit.spacing / 4, 45)
  square = _get_regular_polygon_from_one_side(
    tiling_utils.get_sides(star)[0], 4)
  squares = [affine.rotate(square, a, (0, 0)) for a in range(0, 135, 45)]
  polys = [affine.rotate(p, -22.5, (0, 0)) for p in [star, *squares]]
  polys[2] = affine.scale(
    polys[2], 2, 2, origin = tiling_utils.get_corners(polys[2])[1])
  unit.union = geom.Polygon(
    polys[1].exterior.coords[3:1:-1] + polys[2].exterior.coords[4:1:-1] +
    polys[3].exterior.coords[4:1:-1] + polys[0].exterior.coords[7:])
  vs = tiling_utils.get_corners(polys[0])
  p0, p1 = vs[10], vs[0]
  v1 = geom.Point(p1.x - p0.x, p1.y - p0.y)
  v2 = affine.rotate(v1, 90, (0, 0))
  unit.setup_vectors((v1.x, v1.y), (v2.x, v2.y))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list("abcd")},
    geometry = gpd.GeoSeries(polys),
    crs = unit.crs)


def _setup_chavey(unit:TileUnit) -> str|None:
  """Set up a selection of k-regular tilings by polygons.

  Tilings are from:

  Chavey DP. 1989. Tilings by regular polygons—II: A catalog of tilings.
    Computers & Mathematics with Applications. 17(1-3) 149-165.
    https://doi.org/10.1016/0898-1221(89)90156-9.

  The 'code' letters (alphabetical) are entirely made up and arbitrary as the
  vertex degree based codes offered by Chavey are impossible to make sensible
  use of. See https://www.biglist-tilings.com/ for another approach, which it
  would be interesting to see developed into something a little more coherent!

  NOTE: Not sure that these should be called 'chavey' - more of a grab bag for
  a bunch of other tilings. Unfortunately, their proper codes are unwieldy such
  as 3.3.3.4.4;3.3.4.3.4;4.4.4.4.
  """
  if any(k not in unit.__dict__ for k in ["code"]):
    return "Chavey tilings require a letter code to be be supplied."
  match unit.code:
    case "a":
      _setup_chavey_a(unit)
    case "b":
      _setup_chavey_b(unit)
    case "c":
      _setup_chavey_c(unit)
    case "d":
      _setup_chavey_d(unit)
    case "e":
      _setup_chavey_e(unit)
    case "f":
      _setup_chavey_f(unit)
    case "g":
      _setup_chavey_g(unit)
    case "h":
      _setup_chavey_h(unit)
    case "i":
      _setup_chavey_i(unit)
    case "j":
      _setup_chavey_j(unit)
    case "k":
      _setup_chavey_k(unit)
    case _:
      unit.base_shape = TileShape.HEXAGON
      return "No valid Chavey tiling code supplied."
  return None


def _setup_chavey_a(unit:TileUnit) -> str|None:
  """Set up 3.3.3.4.4;3.3.4.3.4;4.4.4.4."""
  s = unit.spacing / np.sqrt(5 + 2 * sqrt3)
  t = s * sqrt3 / 2
  sq = tiling_utils.get_regular_polygon(s, 4)
  squares = [affine.translate(sq, dx/2, dy/2) for dx in [-s, s]
                                              for dy in [-s, s]]
  squares.append(affine.rotate(squares[0], 150, origin = (-s, -s)))
  tri = tiling_utils.get_regular_polygon(t, 3)
  triangles = [affine.translate(tri, -s/2, s - tri.bounds[1])]
  triangles.append(affine.translate(tri, s/2, s - tri.bounds[1]))
  triangles.extend([affine.rotate(tr, 180, (s/4, s + t/2))
                    for tr in triangles])
  triangles.extend([affine.rotate(tr, 90, (0, 0)) for tr in triangles])
  triangles = [affine.rotate(tr, 90, (0, 0)) for tr in triangles]
  polys = [affine.translate(p, s / sqrt2 * np.cos(np.pi/12),
                               s / sqrt2 * np.sin(np.pi/12))
                            for p in squares + triangles]
  unit.setup_vectors((s/2, 2 * s + t), (2 * s + t, -s/2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:len(polys)]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_chavey_b(unit:TileUnit) -> str|None:
  """Set up 3.4.4.6;3.4.6.4;4.4.4.4."""
  h = unit.spacing * sqrt3 / (1 + sqrt3) / sqrt2
  s = h / sqrt3
  t = h / 2
  hexes_and_tris = [tiling_utils.get_regular_polygon(h, 6)]
  hexes_and_tris.append(geom.Polygon([(-s/2, -t), (s/2, -t), (0, -2*t)]))
  hexes_and_tris.append(affine.rotate(hexes_and_tris[-1], 180, (0, 0)))
  hexes_and_tris.extend([
    affine.rotate(affine.translate(p, 0, h + s), 90, (0, h + s))
    for p in hexes_and_tris])
  squares = [geom.Polygon([(-s, 0), (-s - t, s/2),
                           (-t - s/2, t + s/2), (-s/2, t)])]
  squares.append(affine.translate(squares[0], s/2, t))
  squares.extend([affine.translate(sq, -t, s/2) for sq in squares])
  squares.extend([affine.rotate(sq, 60, (0, h + 2 * s))
                  for sq in squares])
  polys = [affine.translate(p, 0, -s) for p in hexes_and_tris + squares]
  unit.setup_vectors((s + 2 * t,  s + 2 * t),
                     (s + 2 * t, -s - 2 * t))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:len(polys)]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))
  return None


def _setup_chavey_c(unit:TileUnit) -> str|None:
  """Set up 3.4.4.6;3.4.6.4."""
  h = unit.spacing * sqrt3 / (3 + sqrt3)
  s = h / sqrt3
  t = h / 2
  hexes, squares, triangles = [], [], []
  squares.append(tiling_utils.get_regular_polygon(s, 4))
  squares.append(affine.translate(squares[0], 0,  s))
  squares.append(affine.translate(squares[0], 0, -s))
  squares.extend([affine.rotate(sq,  60, origin = (0, t + 3 * s / 2))
                  for sq in squares[:2]])
  squares.extend([affine.rotate(sq, -60, origin = (0, t + 3 * s / 2))
                  for sq in squares[:2]])
  squares.extend([affine.rotate(squares[1], a, origin = (0, t + 3 * s / 2))
                  for a in (120, -120)])
  hexes.append(tiling_utils.get_regular_polygon(h, 6))
  hexes[0] = affine.translate(hexes[0], 0, t + 3 * s / 2)
  hexes.append(affine.rotate(hexes[0], -90, origin = (-s / 2, s / 2)))
  hexes.append(affine.rotate(hexes[1], 180, origin = (0, 0)))
  triangles.append(geom.Polygon([(s/2,    s/2), (s/2, 3*s/2), (s/2 + t, s)]))
  triangles.append(geom.Polygon([(s/2, -3*s/2), (s/2,  -s/2), (s/2 + t, -s)]))
  triangles.extend([affine.rotate(tr, 180, origin = (0, 0))
                    for tr in triangles])
  triangles.extend([affine.rotate(triangles[0], a, origin = (0, t + 3 * s / 2))
                    for a in (60, -120)])
  polys = [affine.translate(p, 0, -t) for p in hexes + squares + triangles]
  unit.setup_vectors((0, h + 3 * s),
                     (3 * (h + s) / 2,  3 * s / 2 + t),
                     (3 * (h + s) / 2, -3 * s / 2 - t))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:len(polys)]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_chavey_d(unit:TileUnit) -> str|None:
  """Set up 3.3.3.4.4;3.3.4.3.4."""
  s = unit.spacing * sqrt2 / (3 + sqrt3)
  t = s * sqrt3 / 2
  squares, triangles = [], []
  squares.append(tiling_utils.get_regular_polygon(s, 4))
  squares.append(affine.translate(squares[0],    0 , -s))
  squares.append(affine.translate(squares[0], -s/2 , t + s))
  squares.append(affine.translate(squares[0],  s/2 , t + s))
  squares.append(affine.rotate(squares[0],  150, ( s/2, s/2)))
  squares.append(affine.rotate(squares[0], -150, (-s/2, s/2)))
  triangles.append(geom.Polygon([(-s/2, -s/2), (-s/2 - t,  0), (-s/2,  s/2)]))
  triangles.append(geom.Polygon([( s/2, -s/2), ( s/2 + t,  0), ( s/2,  s/2)]))
  triangles.append(geom.Polygon([(-s/2, s/2), (-s, t + s/2), ( 0, t + s/2)]))
  triangles.append(geom.Polygon([(-s/2, s/2), ( 0, t + s/2), ( s/2, s/2)]))
  triangles.append(geom.Polygon([( s/2, s/2), ( 0, t + s/2), ( s, t + s/2)]))
  triangles.extend([affine.rotate(tr, 180, (0, s + t))
                    for tr in triangles[2:5]])
  triangles.extend([affine.rotate(triangles[2], a, (-s, t + s/2))
                    for a in (-150, -210)])
  triangles.extend([affine.rotate(triangles[4], a, ( s, t + s/2))
                    for a in ( 150,  210)])
  polys = [affine.translate(p, 0, -t) for p in squares + triangles]
  unit.setup_vectors((unit.spacing / sqrt2,  unit.spacing / sqrt2),
                     (unit.spacing / sqrt2, -unit.spacing / sqrt2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:len(polys)]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_chavey_e(unit:TileUnit) -> str|None:
  """Set up 3.3.3.3.3.3;3.3.4.3.4."""
  s = unit.spacing / (1 + sqrt3)
  t = s * sqrt3 / 2
  squares, triangles = [], []
  triangles.append(geom.Polygon([(0, 0), (s/2, -t), (-s/2, -t)]))
  triangles.extend([affine.rotate(triangles[0], a, (0, 0))
                    for a in range(60, 360, 60)])
  triangles.append(geom.Polygon([(-s/2    ,       t),
                                 (-s/2 - t, s/2 + t),
                                 (-s/2    , s   + t)]))
  triangles.append(geom.Polygon([( s/2    ,        t),
                                 ( s/2    ,  s   + t),
                                 ( s/2 + t,  s/2 + t)]))
  squares.append(affine.translate(
    tiling_utils.get_regular_polygon(s, 4), 0, t + s/2))
  squares.append(affine.rotate(squares[0],  60, (0, 0)))
  squares.append(affine.rotate(squares[0], -60, (0, 0)))
  polys = [affine.translate(p, 0, 0) for p in squares + triangles]
  unit.setup_vectors((0, unit.spacing),
                     (unit.spacing * sqrt3/2, unit.spacing / 2),
                     (unit.spacing * sqrt3/2, -unit.spacing / 2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:len(polys)]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_chavey_f(unit:TileUnit) -> str|None:
  """Set up 3.3.3.3.3.3;3.3.3.3.6."""
  h = unit.spacing * sqrt3 / np.sqrt(7)
  s = h / sqrt3
  t = h / 2
  hexes, triangles = [], []
  hexes.append(tiling_utils.get_regular_polygon(h, 6))
  tri = tiling_utils.get_regular_polygon(t, 3)
  triangles.append(affine.translate(tri, s, -t - tri.bounds[1]))
  triangles.append(affine.rotate(triangles[0], 180, (5*s/4, -t/2)))
  triangles.append(affine.rotate(triangles[1], 180, (3*s/2, 0)))
  triangles.append(affine.rotate(triangles[2], 180, (5*s/4,  t/2)))
  triangles.extend([affine.rotate(tr, 180, (0, 0)) for tr in triangles])
  triangles.extend([affine.rotate(triangles[0], a, (0, 0))
                    for a in (120, 300)])
  polys = hexes + triangles
  unit.setup_vectors((2*s, h), (2*s, -h))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:len(polys)]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))


def _setup_chavey_g(unit:TileUnit) -> str|None:
  """Set up 3.3.3.3.3.3;3.3.3.3.6 (another one)."""
  h = unit.spacing / 2
  s = h / sqrt3
  t = h / 2
  hexes, triangles = [], []
  hexes.append(tiling_utils.get_regular_polygon(h, 6))
  tri = tiling_utils.get_regular_polygon(t, 3)
  triangles.append(affine.translate(tri, s, -t - tri.bounds[1]))
  triangles.append(affine.rotate(triangles[0], 180, (5*s/4, -t/2)))
  triangles.append(affine.rotate(triangles[1], 180, (3*s/2, 0)))
  triangles.extend([affine.rotate(tr, a, (0, 0)) for a in range(60, 360, 60)
                                                 for tr in triangles])
  polys = hexes + triangles
  unit.setup_vectors((0, 2*h), (3*s, h), (3*s, -h))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:len(polys)]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))
  return None

def _setup_chavey_h(unit:TileUnit) -> str|None:
  """Set up 3.3.4.3.4;3.4.6.4."""
  h = unit.spacing / (1 + 2 / sqrt3)
  s = h / sqrt3
  t = h / 2
  hexes, squares, triangles = [], [], []
  hexes.append(tiling_utils.get_regular_polygon(h, 6))
  squares.append(
    affine.translate(tiling_utils.get_regular_polygon(s, 4), 0, t + s/2))
  squares.extend([affine.rotate(squares[0], a, (0, 0))
                  for a in range(60, 360, 60)])
  polys = [affine.rotate(p, 30, (0, 0)) for p in hexes + squares]
  triangles.append(geom.Polygon([(-s/2, -s - t), (0, -s), (s/2, -s - t)]))
  triangles.extend([affine.rotate(triangles[0], a, (0, 0))
                    for a in range(60, 360, 60)])
  triangles.append(geom.Polygon(
                     [(s/2, s + t), (s/2 + t, 3*s/2 + t), (s/2 + t, s/2 + t)]))
  triangles.append(affine.rotate(triangles[-1], 180, (0, 0)))
  polys = [affine.rotate(p, 30, (0, 0)) for p in polys + triangles]
  unit.setup_vectors(( unit.spacing / 2, unit.spacing * sqrt3/2),
                     ( unit.spacing    , 0),
                     (-unit.spacing / 2, unit.spacing * sqrt3/2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:len(polys)]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))
  return None

def _setup_chavey_i(unit:TileUnit) -> str|None:
  """Set up 3.3.3.3.3.3;3.3.4.3.4."""
  h = unit.spacing / (1 + 2 / sqrt3)
  s = h / sqrt3
  t = h / 2
  triangles1, squares, triangles2 = [], [], []
  hex_c = tiling_utils.get_corners(tiling_utils.get_regular_polygon(h, 6))
  triangles1 = [geom.Polygon([(0, 0), p1, p2])
                for p1, p2 in zip(hex_c[:-1], hex_c[1:], strict = True)]
  squares.append(
    affine.translate(tiling_utils.get_regular_polygon(s, 4), 0, t + s/2))
  squares.extend([affine.rotate(squares[0], a, (0, 0))
                  for a in range(60, 360, 60)])
  polys = [affine.rotate(p, 30, (0, 0)) for p in triangles1 + squares]
  triangles2.append(geom.Polygon([(-s/2, -s - t), (0, -s), (s/2, -s - t)]))
  triangles2.extend([affine.rotate(triangles2[0], a, (0, 0))
                     for a in range(60, 360, 60)])
  triangles2.append(geom.Polygon([(s/2    ,   s   + t),
                                  (s/2 + t, 3*s/2 + t),
                                  (s/2 + t,   s/2 + t)]))
  triangles2.append(affine.rotate(triangles2[-1], 180, (0, 0)))
  polys = [affine.rotate(p, 30, (0, 0)) for p in polys + triangles2]
  unit.setup_vectors(( unit.spacing / 2, unit.spacing * sqrt3/2),
                     ( unit.spacing    , 0),
                     (-unit.spacing / 2, unit.spacing * sqrt3/2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:len(polys)]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))
  return None


def _setup_chavey_j(unit:TileUnit) -> str|None:
  h = unit.spacing / (1 + 2 / sqrt3)
  s = h / sqrt3
  t = h / 2
  hexagon = tiling_utils.get_regular_polygon(h, 6)
  sq = affine.translate(tiling_utils.get_regular_polygon(s, 4), 0, (s + h)/2)
  squares = [affine.rotate(sq, a, (0, 0)) for a in range(0, 360, 60)]
  tri = _get_regular_polygon_from_one_side(tiling_utils.get_sides(sq)[0], 3)
  triangles = [affine.rotate(tri, a, (0, 0)) for a in range(0, 360, 60)]
  triangles.append(affine.translate(triangles[3],  s + t,  h + 3*s/2))
  triangles.append(affine.translate(triangles[0], -s - t, -h - 3*s/2))
  polys = [affine.rotate(p, 30, (0, 0))
           for p in [hexagon, *squares, *triangles]]
  unit.setup_vectors(( unit.spacing / 2, unit.spacing * sqrt3/2),
                     ( unit.spacing    , 0),
                     (-unit.spacing / 2, unit.spacing * sqrt3/2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:len(polys)]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))
  return None


def _setup_chavey_k(unit:TileUnit) -> str|None:
  h = unit.spacing / (1 + 2 / sqrt3)
  s = h / sqrt3
  t = h / 2
  triangles1, squares, triangles2 = [], [], []
  hex_c = tiling_utils.get_corners(tiling_utils.get_regular_polygon(h, 6))
  triangles1 = [geom.Polygon([(0, 0), p1, p2])
                for p1, p2 in zip(hex_c[:-1], hex_c[1:], strict = True)]
  squares.append(
    affine.translate(tiling_utils.get_regular_polygon(s, 4), 0, t + s/2))
  squares.extend([affine.rotate(squares[0], a, (0, 0))
                  for a in range(60, 360, 60)])
  polys = [affine.rotate(p, 30, (0, 0)) for p in triangles1 + squares]
  triangles2.append(geom.Polygon([(-s/2, -s - t), (0, -s), (s/2, -s - t)]))
  triangles2.extend([affine.rotate(triangles2[0], a, (0, 0))
                     for a in range(60, 360, 60)])
  triangles2.append(geom.Polygon([(s/2    ,   s   + t),
                                  (s/2 + t, 3*s/2 + t),
                                  (s/2 + t,   s/2 + t)]))
  triangles2.append(affine.rotate(triangles2[-1], 180, (0, 0)))
  polys = polys + triangles2
  polys[-2:] = [affine.rotate(p, 30, (0, 0)) for p in polys[-2:]]
  unit.setup_vectors(( unit.spacing / 2, unit.spacing * sqrt3/2),
                     ( unit.spacing    , 0),
                     (-unit.spacing / 2, unit.spacing * sqrt3/2))
  unit.tiles = gpd.GeoDataFrame(
    data = {"tile_id": list(string.ascii_letters)[:len(polys)]},
    crs = unit.crs,
    geometry = gpd.GeoSeries(polys))
  return None
